package vboxService;
# Generated by SOAP::Lite (v1.27) for Perl -- soaplite.com
# Copyright (C) 2000-2006 Paul Kulchenko, Byrne Reese
#
# CAUTION: This file has been generated using stubmaker.pl from the WSDL file supplied with
# the VirtualBox SDK. Additional manual modifications have also been made.
#
# Customized Methods
# IMedium_createBaseStorage - Allow 5 MediumVariants
# IAppliance_importMachines - Allow 3 ImportOptions
# IAppliance_write - Allow 4 ExportOptions
# IMachine_cloneTo - Allow 4 CloneOptions

# Override the original methods

my %custom_methods = (
'IMedium_createBaseStorage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logicalSize', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
    ], # end parameters
  },
'IMedium_createBaseStorage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logicalSize', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
    ], # end parameters
  }, # end IMedium_createBaseStorage
'IAppliance_importMachines' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ImportOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ImportOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ImportOptions', attr => {}),
    ], # end parameters
  }, # end IAppliance_importMachines
'IAppliance_write' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ExportOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ExportOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ExportOptions', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_write
'IMachine_cloneTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'vbox:CloneMode', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:CloneOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:CloneOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:CloneOptions', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:CloneOptions', attr => {}),
    ], # end parameters
  }, # end IMachine_cloneTo
);



my %methods = (
'IBIOSSettings_getAPICMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getAPICMode
'IConsole_setUseHostClipboard' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'useHostClipboard', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IConsole_setUseHostClipboard
'ICloudNetwork_getProfile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_getProfile
'IVirtualSystemDescription_getValuesByType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:VirtualSystemDescriptionType', attr => {}),
      SOAP::Data->new(name => 'which', type => 'vbox:VirtualSystemDescriptionValueType', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_getValuesByType
'IMedium_setType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:MediumType', attr => {}),
    ], # end parameters
  }, # end IMedium_setType
'IMedium_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getId
'IDisplay_viewportChanged' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_viewportChanged
'ICursorPositionChangedEvent_getX' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICursorPositionChangedEvent_getX
'INetworkAdapter_getBridgedInterface' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getBridgedInterface
'IEvent_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEvent_getType
'IMedium_getHostDrive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getHostDrive
'IHost_getProcessorOnlineCoreCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorOnlineCoreCount
'IReusableEvent_reuse' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IReusableEvent_reuse
'ICertificate_getExpired' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getExpired
'IMachine_getGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getGroups
'IMachine_getSnapshotCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSnapshotCount
'IMachine_setChipsetType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'chipsetType', type => 'vbox:ChipsetType', attr => {}),
    ], # end parameters
  }, # end IMachine_setChipsetType
'IMachine_getExtraDataKeys' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getExtraDataKeys
'IConsole_getGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getGuest
'IMachine_setSettingsFilePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settingsFilePath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setSettingsFilePath
'IRangedIntegerFormValue_getMinimum' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRangedIntegerFormValue_getMinimum
'IMachine_getKeyboardHIDType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getKeyboardHIDType
'IGuestFileStateChangedEvent_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileStateChangedEvent_getStatus
'IMachine_querySavedGuestScreenInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_querySavedGuestScreenInfo
'INetworkAdapter_getLineSpeed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getLineSpeed
'IMachine_setHotPluggableForDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'hotPluggable', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setHotPluggableForDevice
'ISerialPort_setHostMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostMode', type => 'vbox:PortMode', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setHostMode
'IDHCPConfig_remove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_remove
'IFramebufferOverlay_setVisible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'visible', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_setVisible
'IHost_getProcessorFeature' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'feature', type => 'vbox:ProcessorFeature', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorFeature
'IBandwidthGroupChangedEvent_getBandwidthGroup' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthGroupChangedEvent_getBandwidthGroup
'ISnapshotRestoredEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshotRestoredEvent_getMidlDoesNotLikeEmptyInterfaces
'IHostNetworkInterface_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getId
'IMachineDebugger_getPATMEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getPATMEnabled
'IBIOSSettings_getPXEDebugEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getPXEDebugEnabled
'ISystemProperties_setFreeDiskSpacePercentError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'freeDiskSpacePercentError', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setFreeDiskSpacePercentError
'IHostNetworkInterface_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getName
'IUnattended_getLocale' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getLocale
'IGuestSession_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getName
'IUSBDevice_getVendorId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getVendorId
'IConsole_getSharedFolders' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getSharedFolders
'IMedium_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getDescription
'IMachine_setVMProcessPriority' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'VMProcessPriority', type => 'vbox:VMProcPriority', attr => {}),
    ], # end parameters
  }, # end IMachine_setVMProcessPriority
'IGuest_internalGetStatistics' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_internalGetStatistics
'IMachine_detachHostPCIDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostAddress', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMachine_detachHostPCIDevice
'IFsObjInfo_getAccessTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getAccessTime
'IConsole_getDebugger' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getDebugger
'IStorageDeviceChangedEvent_getRemoved' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageDeviceChangedEvent_getRemoved
'IFramebufferOverlay_getVisible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_getVisible
'IMachine_setCPUProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'vbox:CPUPropertyType', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUProperty
'IVFSExplorer_entryList' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_entryList
'INATNetworkPortForwardEvent_getHostPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getHostPort
'IMachineDebugger_writeVirtualMemory' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'address', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bytes', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_writeVirtualMemory
'IChoiceFormValue_getValues' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IChoiceFormValue_getValues
'IMachineDebugger_getOSName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getOSName
'IDHCPConfig_setSuppressedOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'suppressedOptions', type => 'vbox:DHCPOption', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_setSuppressedOptions
'IPerformanceMetric_getCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getCount
'IGraphicsAdapter_getAccelerate3DEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_getAccelerate3DEnabled
'IVRDEServer_getVRDEExtPack' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getVRDEExtPack
'IGuestScreenInfo_getOrigin' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getOrigin
'ICloudClient_importImage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'parameters', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_importImage
'IMachineDebugger_unloadPlugIn' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_unloadPlugIn
'ICloudProvider_getPropertyDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getPropertyDescription
'IGuestMultiTouchEvent_getYPositions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMultiTouchEvent_getYPositions
'INATNetwork_addPortForwardRule' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'isIpv6', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'ruleName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'proto', type => 'vbox:NATProtocol', attr => {}),
      SOAP::Data->new(name => 'hostIP', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostPort', type => 'xsd:unsignedShort', attr => {}),
      SOAP::Data->new(name => 'guestIP', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'guestPort', type => 'xsd:unsignedShort', attr => {}),
    ], # end parameters
  }, # end INATNetwork_addPortForwardRule
'IUnattended_getFullUserName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getFullUserName
'IFramebufferOverlay_getY' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_getY
'IMachine_getAutostartDelay' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAutostartDelay
'IVirtualSystemDescription_setFinalValues' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'VBoxValues', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'extraConfigValues', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_setFinalValues
'IDHCPGroupCondition_setInclusive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'inclusive', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_setInclusive
'IBIOSSettings_getLogoDisplayTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getLogoDisplayTime
'IFramebuffer_getBitsPerPixel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getBitsPerPixel
'ISystemProperties_setProxyURL' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'proxyURL', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setProxyURL
'IUSBControllerChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBControllerChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'IFormValue_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getEnabled
'IMachine_getDefaultFrontend' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getDefaultFrontend
'IGuestSession_environmentScheduleUnset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_environmentScheduleUnset
'IFramebuffer_getOverlay' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getOverlay
'IFile_close' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_close
'IBIOSSettings_setAPICMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'APICMode', type => 'vbox:APICMode', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setAPICMode
'IUnattended_getUser' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getUser
'INATNetwork_setIPv6Prefix' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IPv6Prefix', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setIPv6Prefix
'ISystemProperties_getLoggingLevel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getLoggingLevel
'IEmulatedUSB_webcamAttach' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settings', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEmulatedUSB_webcamAttach
'IVirtualSystemDescription_addDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:VirtualSystemDescriptionType', attr => {}),
      SOAP::Data->new(name => 'VBoxValue', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'extraConfigValue', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_addDescription
'IMachineDebugger_setSingleStep' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'singleStep', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setSingleStep
'IHost_getUSBDevices' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getUSBDevices
'IGuestSession_getPathStyle' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getPathStyle
'IGuestPropertyChangedEvent_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestPropertyChangedEvent_getName
'IMachine_setOSTypeId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'OSTypeId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setOSTypeId
'IHostVideoInputDevice_getPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostVideoInputDevice_getPath
'IVRDEServer_setAuthLibrary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'authLibrary', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setAuthLibrary
'IMachine_queryLogFilename' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'idx', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_queryLogFilename
'IMousePointerShapeChangedEvent_getWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getWidth
'INATNetworkChangedEvent_getNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkChangedEvent_getNetworkName
'IAppliance_addPasswords' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'identifiers', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'passwords', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_addPasswords
'IMachine_getStorageControllerByInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'connectionType', type => 'vbox:StorageBus', attr => {}),
      SOAP::Data->new(name => 'instance', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getStorageControllerByInstance
'IGuestSessionRegisteredEvent_getRegistered' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSessionRegisteredEvent_getRegistered
'IHost_getNetworkInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getNetworkInterfaces
'IVRDEServer_getReuseSingleConnection' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getReuseSingleConnection
'IMachineEvent_getMachineId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineEvent_getMachineId
'IRecordingScreenSettings_getVideoFPS' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoFPS
'ISharedFolder_getWritable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getWritable
'IKeyboard_putScancodes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'scancodes', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IKeyboard_putScancodes
'IRecordingSettings_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IRecordingSettings_setEnabled
'IFramebufferOverlay_getAlpha' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_getAlpha
'IVirtualBox_composeMachineFilename' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'group', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'createFlags', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'baseFolder', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_composeMachineFilename
'IClipboardFileTransferModeChangedEvent_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IClipboardFileTransferModeChangedEvent_getEnabled
'INATNetwork_getGateway' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getGateway
'IGuestSession_directoryRemoveRecursive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:DirectoryRemoveRecFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryRemoveRecursive
'IDHCPGroupCondition_setType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:DHCPGroupConditionType', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_setType
'IGraphicsAdapter_setMonitorCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'monitorCount', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_setMonitorCount
'IDnDTarget_move' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'defaultAction', type => 'vbox:DnDAction', attr => {}),
      SOAP::Data->new(name => 'allowedActions', type => 'vbox:DnDAction', attr => {}),
      SOAP::Data->new(name => 'formats', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDTarget_move
'IGuestProcessStateChangedEvent_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessStateChangedEvent_getStatus
'IDHCPServer_getIndividualConfigs' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getIndividualConfigs
'IMachine_setTeleporterPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'teleporterPort', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setTeleporterPort
'IDHCPServer_getUpperIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getUpperIP
'IAppliance_getDisks' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getDisks
'IMachine_getSettingsAuxFilePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSettingsAuxFilePath
'IGuestSession_environmentDoesBaseVariableExist' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_environmentDoesBaseVariableExist
'IDnDTarget_enter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'defaultAction', type => 'vbox:DnDAction', attr => {}),
      SOAP::Data->new(name => 'allowedActions', type => 'vbox:DnDAction', attr => {}),
      SOAP::Data->new(name => 'formats', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDTarget_enter
'ISystemProperties_getMediumFormats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMediumFormats
'ICloudClient_getImportDescriptionForm' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_getImportDescriptionForm
'IFsObjInfo_getGroupName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getGroupName
'IMedium_getAllowedTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getAllowedTypes
'IStorageController_getBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getBus
'IHost_getProcessorDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorDescription
'IMedium_setIds' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'setImageId', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'imageId', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'setParentId', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'parentId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_setIds
'IEventSource_getEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'listener', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IEventSource_getEvent
'IGuestSession_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getId
'IUSBDeviceFilter_getRemote' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getRemote
'IUnattended_getExtraInstallKernelParameters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getExtraInstallKernelParameters
'IFile_read' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'toRead', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFile_read
'IVFSExplorer_cdUp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_cdUp
'ISystemProperties_getDeviceTypesForStorageBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bus', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDeviceTypesForStorageBus
'IProgress_getTimeRemaining' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getTimeRemaining
'IGuestSession_processGet' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pid', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_processGet
'IGuestSession_directoryOpen' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'filter', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:DirectoryOpenFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryOpen
'IMachine_getSharedFolders' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSharedFolders
'IAudioAdapter_getEnabledOut' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getEnabledOut
'INATEngine_getDNSPassDomain' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getDNSPassDomain
'IStringFormValue_getString' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStringFormValue_getString
'ICloudProviderManager_getProviders' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProviderManager_getProviders
'IDnDBase_addFormats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'formats', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDBase_addFormats
'IGuestAdditionsStatusChangedEvent_getRunLevel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestAdditionsStatusChangedEvent_getRunLevel
'IConsole_powerUp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_powerUp
'IUnattended_getDetectedOSFlavor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getDetectedOSFlavor
'IConsole_pause' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_pause
'IMachine_getParent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getParent
'ISerialPort_setServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'server', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setServer
'INetworkAdapter_setHostOnlyInterface' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostOnlyInterface', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setHostOnlyInterface
'IMachineDebugger_modifyLogGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settings', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_modifyLogGroups
'IUnattended_setInstallTestExecService' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'installTestExecService', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IUnattended_setInstallTestExecService
'ICertificate_getSignatureAlgorithmOID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSignatureAlgorithmOID
'IGuestSession_setEnvironmentChanges' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'environmentChanges', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_setEnvironmentChanges
'ISystemProperties_getFreeDiskSpacePercentWarning' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getFreeDiskSpacePercentWarning
'IMousePointerShapeChangedEvent_getShape' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getShape
'IMachineDebugger_queryOSKernelLog' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maxMessages', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_queryOSKernelLog
'IHost_getProcessorCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorCount
'IMedium_cloneTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'parent', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_cloneTo
'ICanShowWindowEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICanShowWindowEvent_getMidlDoesNotLikeEmptyInterfaces
'IGuestSession_getDirectories' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getDirectories
'IGraphicsAdapter_getGraphicsControllerType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_getGraphicsControllerType
'ISnapshot_getMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getMachine
'IMedium_deleteStorage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_deleteStorage
'IDHCPGroupConfig_getConditions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupConfig_getConditions
'IMachineDebugger_getRegister' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getRegister
'IKeyboardLedsChangedEvent_getNumLock' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboardLedsChangedEvent_getNumLock
'IDHCPGroupCondition_getValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_getValue
'IUnattended_setProxy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'proxy', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setProxy
'IDnDBase_isFormatSupported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDBase_isFormatSupported
'IMachineDebugger_setExecuteAllInIEM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'executeAllInIEM', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setExecuteAllInIEM
'IDHCPConfig_removeOption' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'option', type => 'vbox:DHCPOption', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_removeOption
'ISystemProperties_getMinGuestCPUCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMinGuestCPUCount
'ISystemProperties_setLogHistoryCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logHistoryCount', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setLogHistoryCount
'IMachine_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getState
'ICloudProfile_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_getName
'IEvent_getSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEvent_getSource
'IChoiceFormValue_setSelectedIndex' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'index', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IChoiceFormValue_setSelectedIndex
'IVirtualBox_getGuestOSTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getGuestOSTypes
'IGuestProcessEvent_getPid' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessEvent_getPid
'INATNetworkStartStopEvent_getStartEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkStartStopEvent_getStartEvent
'IConsole_addDiskEncryptionPasswords' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'ids', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'passwords', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'clearOnSuspend', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IConsole_addDiskEncryptionPasswords
'IRecordingScreenSettings_isFeatureEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'feature', type => 'vbox:RecordingFeature', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_isFeatureEnabled
'IRecordingScreenSettings_getAudioCodec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getAudioCodec
'IMedium_getProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getProperty
'IRecordingScreenSettings_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getId
'IVFSExplorer_cd' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'dir', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_cd
'IGraphicsAdapter_setAccelerate2DVideoEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'accelerate2DVideoEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_setAccelerate2DVideoEnabled
'ICloudNetwork_setProvider' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'provider', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_setProvider
'IMachine_setAllowTracingToAccessVM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'allowTracingToAccessVM', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setAllowTracingToAccessVM
'IMedium_getFormat' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getFormat
'IMedium_setProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_setProperty
'ISystemProperties_getSupportedNetworkAdapterTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedNetworkAdapterTypes
'IDHCPServer_findLeaseByMAC' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mac', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_findLeaseByMAC
'IFile_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getStatus
'IProgress_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getId
'ISession_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_getState
'ISerialPort_getPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getPath
'IHost_removeUSBDeviceFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'position', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IHost_removeUSBDeviceFilter
'IPerformanceMetric_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getDescription
'IProgressPercentageChangedEvent_getPercent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgressPercentageChangedEvent_getPercent
'IGuestSession_directoryCopy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:DirectoryCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryCopy
'IPCIAddress_getDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_getDevice
'IDHCPServer_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getEnabled
'IHost_createUSBDeviceFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_createUSBDeviceFilter
'IMachine_getCPUProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'vbox:CPUPropertyType', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUProperty
'IBIOSSettings_setLogoFadeOut' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logoFadeOut', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setLogoFadeOut
'IGuestSessionEvent_getSession' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSessionEvent_getSession
'ISystemProperties_getMaxInstancesOfUSBControllerType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'chipset', type => 'vbox:ChipsetType', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:USBControllerType', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxInstancesOfUSBControllerType
'ICloudNetwork_setNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_setNetworkName
'IStorageController_setUseHostIOCache' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'useHostIOCache', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IStorageController_setUseHostIOCache
'IHost_getNameServers' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getNameServers
'IUnattended_getIsoPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getIsoPath
'INATNetwork_getIPv6Prefix' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getIPv6Prefix
'IMedium_getDeviceType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getDeviceType
'INetworkAdapter_setLineSpeed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'lineSpeed', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setLineSpeed
'INATEngine_setTFTPNextServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'TFTPNextServer', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_setTFTPNextServer
'IHostNetworkInterface_getIPV6Supported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getIPV6Supported
'IExtraDataChangedEvent_getValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IExtraDataChangedEvent_getValue
'IVRDEServer_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setEnabled
'IMachine_setAutostartEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autostartEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setAutostartEnabled
'IMachine_deleteConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'media', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_deleteConfig
'IRecordingScreenSettings_getFeatures' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getFeatures
'IRecordingScreenSettings_setVideoRate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoRate', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoRate
'ISystemProperties_getDefaultAdditionsISO' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultAdditionsISO
'INATNetwork_removePortForwardRule' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'iSipv6', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'ruleName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_removePortForwardRule
'IMachine_getChipsetType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getChipsetType
'IMachine_setGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'groups', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setGroups
'IConsole_removeSharedFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_removeSharedFolder
'IMachine_adoptSavedState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'savedStateFile', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_adoptSavedState
'ISystemProperties_getScreenShotFormats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getScreenShotFormats
'ICloudClient_startInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_startInstance
'IMachine_getSerialPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'slot', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getSerialPort
'IBandwidthControl_getNumGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthControl_getNumGroups
'IGuestFileReadEvent_getData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileReadEvent_getData
'IMachine_getVRDEServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getVRDEServer
'IStorageController_getUseHostIOCache' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getUseHostIOCache
'IHostNameResolutionConfigurationChangeEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNameResolutionConfigurationChangeEvent_getMidlDoesNotLikeEmptyInterfaces
'IMachine_getUSBControllerByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getUSBControllerByName
'IPerformanceCollector_getMetrics' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'metricNames', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'objects', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceCollector_getMetrics
'INATRedirectEvent_getHostIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getHostIP
'IVetoEvent_getVetos' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVetoEvent_getVetos
'IGuest_getStatisticsUpdateInterval' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getStatisticsUpdateInterval
'IUnattended_getInstallGuestAdditions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getInstallGuestAdditions
'IMediumIO_convertToStream' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
      SOAP::Data->new(name => 'bufferSize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMediumIO_convertToStream
'IVFSExplorer_exists' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_exists
'IRecordingScreenSettings_getDestination' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getDestination
'IMachine_takeSnapshot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pause', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_takeSnapshot
'IUSBDeviceFilter_getManufacturer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getManufacturer
'IVirtualSystemDescription_getCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_getCount
'IMachineDebugger_getHWVirtExNestedPagingEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getHWVirtExNestedPagingEnabled
'IFsObjInfo_getUID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getUID
'IMediumFormat_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumFormat_getId
'IVirtualBox_getRevision' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getRevision
'IProcess_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getEventSource
'IProcess_getExitCode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getExitCode
'IDataStream_getReadSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDataStream_getReadSize
'IDisplay_createGuestScreenInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'display', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'status', type => 'vbox:GuestMonitorStatus', attr => {}),
      SOAP::Data->new(name => 'primary', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'changeOrigin', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'originX', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'originY', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bitsPerPixel', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_createGuestScreenInfo
'IMachine_getEmulatedUSBCardReaderEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getEmulatedUSBCardReaderEnabled
'IUnattended_setFullUserName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'fullUserName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setFullUserName
'ISystemProperties_getVRDEAuthLibrary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getVRDEAuthLibrary
'IAudioAdapterChangedEvent_getAudioAdapter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapterChangedEvent_getAudioAdapter
'IStringFormValue_setString' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'text', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStringFormValue_setString
'IRecordingScreenSettings_setVideoFPS' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoFPS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoFPS
'IMachine_getOSTypeId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getOSTypeId
'IGuestScreenInfo_getHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getHeight
'IUnattended_getAdditionsIsoPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getAdditionsIsoPath
'IProcess_waitFor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'waitFor', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IProcess_waitFor
'ICPUExecutionCapChangedEvent_getExecutionCap' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICPUExecutionCapChangedEvent_getExecutionCap
'IHost_findHostDVDDrive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_findHostDVDDrive
'ICloudProvider_createProfile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'profileName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'values', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_createProfile
'ICloudProvider_getShortName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getShortName
'IMachine_enumerateGuestProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'patterns', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_enumerateGuestProperties
'IVirtualSystemDescription_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_getDescription
'IMachineDebugger_getExecutionEngine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getExecutionEngine
'IAppliance_getWarnings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getWarnings
'IUnattended_setAdditionsIsoPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'additionsIsoPath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setAdditionsIsoPath
'IFramebuffer_getBytesPerLine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getBytesPerLine
'IMediumIO_formatFAT' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'quick', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMediumIO_formatFAT
'IKeyboard_releaseKeys' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboard_releaseKeys
'IGuestSession_symlinkCreate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'symlink', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:SymlinkType', attr => {}),
    ], # end parameters
  }, # end IGuestSession_symlinkCreate
'IGuest_setStatisticsUpdateInterval' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'statisticsUpdateInterval', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuest_setStatisticsUpdateInterval
'IUSBController_getUSBStandard' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBController_getUSBStandard
'IStorageController_setControllerType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerType', type => 'vbox:StorageControllerType', attr => {}),
    ], # end parameters
  }, # end IStorageController_setControllerType
'ICloudClient_getImageInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_getImageInfo
'IMachine_addUSBController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:USBControllerType', attr => {}),
    ], # end parameters
  }, # end IMachine_addUSBController
'IMedium_getLogicalSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getLogicalSize
'IGuestDnDTarget_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestDnDTarget_getMidlDoesNotLikeEmptyInterfaces
'IMachineDebugger_getUptime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getUptime
'IMedium_createDiffStorage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
    ], # end parameters
  }, # end IMedium_createDiffStorage
'IHostNetworkInterface_enableStaticIPConfigV6' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IPV6Address', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IPV6NetworkMaskPrefixLength', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_enableStaticIPConfigV6
'ICloudNetworkGatewayInfo_getInstanceId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetworkGatewayInfo_getInstanceId
'IVirtualBoxErrorInfo_getResultDetail' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBoxErrorInfo_getResultDetail
'IDHCPConfig_setMaxLeaseTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maxLeaseTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_setMaxLeaseTime
'IMachineDebugger_setRegisters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'values', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setRegisters
'IMachine_getTracingConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getTracingConfig
'IVirtualBox_createSharedFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostPath', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'writable', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'automount', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'autoMountPoint', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createSharedFolder
'IMachine_getAudioAdapter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAudioAdapter
'IUnattended_getImageIndex' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getImageIndex
'IGuestSession_fsObjSetACL' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'followSymlinks', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'acl', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjSetACL
'IGuestSession_waitFor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'waitFor', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_waitFor
'IUnattended_setAuxiliaryBasePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'auxiliaryBasePath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setAuxiliaryBasePath
'IMedium_getBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getBase
'ICertificate_getSignatureAlgorithmName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSignatureAlgorithmName
'INATNetwork_getLocalMappings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getLocalMappings
'IRecordingScreenSettings_setVideoWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoWidth', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoWidth
'IReusableEvent_getGeneration' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IReusableEvent_getGeneration
'ICPUChangedEvent_getCPU' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICPUChangedEvent_getCPU
'IProgressEvent_getProgressId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgressEvent_getProgressId
'IMachine_hotPlugCPU' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpu', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_hotPlugCPU
'IStorageController_getInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getInstance
'IRecordingScreenSettings_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setEnabled
'IMachine_getLogFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getLogFolder
'IStorageController_setInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'instance', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IStorageController_setInstance
'IMachine_nonRotationalDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'nonRotational', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_nonRotationalDevice
'ICertificate_getCertificateAuthority' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getCertificateAuthority
'IAudioAdapter_setAudioCodec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioCodec', type => 'vbox:AudioCodecType', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setAudioCodec
'IMachine_getGuestPropertyTimestamp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getGuestPropertyTimestamp
'IUSBDeviceFilter_getMaskedInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getMaskedInterfaces
'IMachine_getStorageControllerByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getStorageControllerByName
'IGuestSession_processCreate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'executable', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'arguments', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'environmentChanges', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:ProcessCreateFlag', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_processCreate
'IFramebufferOverlay_getX' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_getX
'IMachineDebugger_setCSAMEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'CSAMEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setCSAMEnabled
'IConsole_powerDown' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_powerDown
'ISystemProperties_getSupportedStorageBuses' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedStorageBuses
'IVirtualBox_getMachineGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getMachineGroups
'IAudioAdapter_getEnabledIn' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getEnabledIn
'ISystemProperties_getSupportedRecordingVRCModes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedRecordingVRCModes
'IMachine_getUSBProxyAvailable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getUSBProxyAvailable
'IVirtualBox_getMachineStates' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'machines', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getMachineStates
'IMediumFormat_describeProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumFormat_describeProperties
'IAppliance_getVirtualSystemDescriptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getVirtualSystemDescriptions
'IAudioAdapter_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getEnabled
'IDHCPConfig_setDefaultLeaseTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultLeaseTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_setDefaultLeaseTime
'IKeyboard_putUsageCode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'usageCode', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'usagePage', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'keyRelease', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IKeyboard_putUsageCode
'IParallelPort_getPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPort_getPath
'INetworkAdapter_getNATEngine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getNATEngine
'IGuestMonitorChangedEvent_getOriginX' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorChangedEvent_getOriginX
'IMachineDebugger_getLogDbgFlags' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogDbgFlags
'ISnapshot_getParent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getParent
'IUnattended_constructMedia' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_constructMedia
'IDisplay_setScreenLayout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenLayoutMode', type => 'vbox:ScreenLayoutMode', attr => {}),
      SOAP::Data->new(name => 'guestScreenInfo', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDisplay_setScreenLayout
'IAppliance_getPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getPath
'IStringFormValue_getMultiline' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStringFormValue_getMultiline
'IUSBDeviceFilter_getProductId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getProductId
'ISystemProperties_getStorageControllerTypesForStorageBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'storageBus', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getStorageControllerTypesForStorageBus
'IHostPCIDevicePlugEvent_getSuccess' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostPCIDevicePlugEvent_getSuccess
'IMachine_getIOCacheEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getIOCacheEnabled
'IGuestFileOffsetChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileOffsetChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'IMediumFormat_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumFormat_getName
'IMachine_mountMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'medium', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'force', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_mountMedium
'IFsObjInfo_getUserFlags' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getUserFlags
'IProgress_getResultCode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getResultCode
'IVRDEServer_setAllowMultiConnection' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'allowMultiConnection', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setAllowMultiConnection
'ISystemProperties_getSupportedDnDModes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedDnDModes
'ICloudClient_exportVM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'progress', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_exportVM
'ICloudClient_launchVM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_launchVM
'IGuestMultiTouchEvent_getScanTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMultiTouchEvent_getScanTime
'IEventSourceChangedEvent_getAdd' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventSourceChangedEvent_getAdd
'IFsObjInfo_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getType
'INATEngine_setNetworkSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mtu', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'sockSnd', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'sockRcv', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'TcpWndSnd', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'TcpWndRcv', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end INATEngine_setNetworkSettings
'IDnDSource_receiveData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDSource_receiveData
'IMedium_getLastAccessError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getLastAccessError
'IMedium_cloneToBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
    ], # end parameters
  }, # end IMedium_cloneToBase
'IGuestSession_fileCopy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FileCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileCopy
'ICertificate_getSubjectPublicKey' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSubjectPublicKey
'IMedium_setDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_setDescription
'ISystemProperties_getSupportedRecordingVideoCodecs' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedRecordingVideoCodecs
'IUnattended_getPackageSelectionAdjustments' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getPackageSelectionAdjustments
'IHost_getOperatingSystem' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getOperatingSystem
'IHostNetworkInterface_getMediumType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getMediumType
'IProgressTaskCompletedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgressTaskCompletedEvent_getMidlDoesNotLikeEmptyInterfaces
'IUnattended_getValidationKitIsoPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getValidationKitIsoPath
'IDHCPServer_getNetworkMask' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getNetworkMask
'IUnattended_setHostname' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostname', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setHostname
'ISessionStateChangedEvent_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISessionStateChangedEvent_getState
'IGuestSession_fsObjRemove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjRemove
'IMachineDebugger_setRecompileUser' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'recompileUser', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setRecompileUser
'ICertificate_isCurrentlyExpired' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_isCurrentlyExpired
'ICertificate_getTrusted' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getTrusted
'IStorageController_getMinPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getMinPortCount
'IMachine_getCurrentSnapshot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCurrentSnapshot
'IMedium_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getType
'IRecordingScreenSettings_setAudioChannels' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioChannels', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setAudioChannels
'IGuest_getAdditionsRunLevel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getAdditionsRunLevel
'IProcess_getArguments' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getArguments
'IMachine_setDefaultFrontend' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultFrontend', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setDefaultFrontend
'IGuest_getDnDTarget' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getDnDTarget
'IUnattended_getLanguage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getLanguage
'IHost_findUSBDeviceByAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_findUSBDeviceByAddress
'ICloudProvider_importProfiles' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_importProfiles
'IMedium_setAutoReset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autoReset', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMedium_setAutoReset
'IMachineDebugger_loadPlugIn' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_loadPlugIn
'IHostNetworkInterface_getNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getNetworkName
'IMouse_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouse_getEventSource
'IRecordingSettings_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingSettings_getEnabled
'ISystemProperties_getMaxBootPosition' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxBootPosition
'IMachine_getParavirtProvider' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getParavirtProvider
'IGuestPropertyChangedEvent_getValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestPropertyChangedEvent_getValue
'ISystemProperties_getDefaultHardDiskFormat' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultHardDiskFormat
'INetworkAdapter_getBootPriority' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getBootPriority
'ISystemProperties_setDefaultHardDiskFormat' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultHardDiskFormat', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setDefaultHardDiskFormat
'IMachine_getLastStateChange' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getLastStateChange
'INATNetworkAlterEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkAlterEvent_getMidlDoesNotLikeEmptyInterfaces
'ICloudProviderManager_getProviderById' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'providerId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProviderManager_getProviderById
'IMachine_setTeleporterPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'teleporterPassword', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setTeleporterPassword
'ISystemProperties_getInfoVDSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getInfoVDSize
'IGuest_setMemoryBalloonSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'memoryBalloonSize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuest_setMemoryBalloonSize
'IPerformanceCollector_enableMetrics' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'metricNames', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'objects', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceCollector_enableMetrics
'IPCIAddress_setBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bus', type => 'xsd:short', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_setBus
'IVRDEServer_setAuthType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'authType', type => 'vbox:AuthType', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setAuthType
'IGraphicsAdapter_setGraphicsControllerType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'graphicsControllerType', type => 'vbox:GraphicsControllerType', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_setGraphicsControllerType
'ICloudProvider_prepareUninstall' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_prepareUninstall
'IGuestMouseEvent_getMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMouseEvent_getMode
'IProgress_getCompleted' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getCompleted
'IGuestFileIOEvent_getOffset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileIOEvent_getOffset
'IProgress_setTimeout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IProgress_setTimeout
'IDHCPServer_restart' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_restart
'IGuestScreenInfo_getPrimary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getPrimary
'IMedium_getMediumFormat' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getMediumFormat
'IFramebuffer_notifyUpdate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_notifyUpdate
'IMedium_getMachineIds' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getMachineIds
'IMachine_setParavirtProvider' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'paravirtProvider', type => 'vbox:ParavirtProvider', attr => {}),
    ], # end parameters
  }, # end IMachine_setParavirtProvider
'ISnapshot_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getId
'INATNetworkPortForwardEvent_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getName
'IGuestMonitorChangedEvent_getHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorChangedEvent_getHeight
'IGuestMonitorChangedEvent_getScreenId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorChangedEvent_getScreenId
'ISystemProperties_getMaxGuestMonitors' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxGuestMonitors
'ISerialPort_setIRQ' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IRQ', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setIRQ
'IGuestScreenInfo_getExtendedInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getExtendedInfo
'ISystemProperties_getProxyMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getProxyMode
'IPerformanceCollector_queryMetricsData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'metricNames', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'objects', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceCollector_queryMetricsData
'IDHCPServer_getIPAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getIPAddress
'IToken_dummy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IToken_dummy
'ISystemProperties_getSupportedChipsetTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedChipsetTypes
'IMouse_getAbsoluteSupported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouse_getAbsoluteSupported
'IHost_findHostFloppyDrive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_findHostFloppyDrive
'IGuestFileStateChangedEvent_getError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileStateChangedEvent_getError
'IMachineDataChangedEvent_getTemporary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDataChangedEvent_getTemporary
'IDHCPGroupConfig_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupConfig_getName
'ISystemProperties_getSupportedRecordingVSMethods' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedRecordingVSMethods
'IGuestProcessInputNotifyEvent_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessInputNotifyEvent_getStatus
'IMachine_unmountMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'force', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_unmountMedium
'IAppliance_getMachines' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getMachines
'IAppliance_createVirtualSystemDescriptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'requested', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IAppliance_createVirtualSystemDescriptions
'IGuestSession_getProtocolVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getProtocolVersion
'IDHCPConfig_setOption' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'option', type => 'vbox:DHCPOption', attr => {}),
      SOAP::Data->new(name => 'encoding', type => 'vbox:DHCPOptionEncoding', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_setOption
'IMachine_getMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMachine_getMedium
'IMouse_putEventMultiTouchString' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'count', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'contacts', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'scanTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMouse_putEventMultiTouchString
'IMachineDebugger_resetStats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pattern', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_resetStats
'INATEngine_getHostIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getHostIP
'IUnattended_getCountry' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getCountry
'IUSBDevice_getBackend' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getBackend
'INATNetwork_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setEnabled
'IDirectory_getDirectoryName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDirectory_getDirectoryName
'ISystemProperties_getLogHistoryCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getLogHistoryCount
'IMousePointerShape_getAlpha' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getAlpha
'IExtraDataChangedEvent_getKey' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IExtraDataChangedEvent_getKey
'IHost_getMemoryAvailable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getMemoryAvailable
'IMachine_deleteSnapshotRange' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'startId', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'endId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_deleteSnapshotRange
'IFramebuffer_getCapabilities' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getCapabilities
'IFormValue_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getDescription
'IEventListener_handleEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'event', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventListener_handleEvent
'INATNetwork_getNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getNetworkName
'IVirtualBoxErrorInfo_getText' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBoxErrorInfo_getText
'IRecordingScreenSettings_setVideoRateControlMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoRateControlMode', type => 'vbox:RecordingVideoRateControlMode', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoRateControlMode
'IDHCPConfig_setForcedOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'forcedOptions', type => 'vbox:DHCPOption', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_setForcedOptions
'ICloudNetwork_setProfile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'profile', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_setProfile
'IMachine_setMemorySize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'memorySize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setMemorySize
'IMachineDebugger_dumpStats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pattern', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_dumpStats
'IGuestSession_getFiles' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getFiles
'IMedium_getLocation' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getLocation
'IDnDBase_getProtocolVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDBase_getProtocolVersion
'IGuestMultiTouchEvent_getXPositions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMultiTouchEvent_getXPositions
'IMachine_setCPUHotPlugEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'CPUHotPlugEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUHotPlugEnabled
'IGuest_getFacilities' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getFacilities
'IMachine_getCPUExecutionCap' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUExecutionCap
'INATNetwork_setNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'network', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setNetwork
'IMachineDebugger_readPhysicalMemory' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'address', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_readPhysicalMemory
'IMachine_removeStorageController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_removeStorageController
'IGuestProcessOutputEvent_getData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessOutputEvent_getData
'IBIOSSettings_setLogoDisplayTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logoDisplayTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setLogoDisplayTime
'IEvent_getWaitable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEvent_getWaitable
'IKeyboard_putCAD' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboard_putCAD
'IMachine_readLog' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'idx', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IMachine_readLog
'IVirtualBox_getProgressOperations' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getProgressOperations
'IVetoEvent_getApprovals' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVetoEvent_getApprovals
'IMachine_removeUSBController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_removeUSBController
'ICursorPositionChangedEvent_getY' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICursorPositionChangedEvent_getY
'IGuestAdditionsStatusChangedEvent_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestAdditionsStatusChangedEvent_getStatus
'ISystemProperties_getSupportedGraphicsControllerTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedGraphicsControllerTypes
'IDHCPConfig_setMinLeaseTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'minLeaseTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_setMinLeaseTime
'IRecordingScreenSettings_getVideoRateControlMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoRateControlMode
'IFile_seek' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'whence', type => 'vbox:FileSeekOrigin', attr => {}),
    ], # end parameters
  }, # end IFile_seek
'IMachineDebugger_getLogDbgGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogDbgGroups
'IGuestSessionStateChangedEvent_getError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSessionStateChangedEvent_getError
'IGuestSession_directoryExists' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'followSymlinks', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryExists
'IConsole_findUSBDeviceById' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_findUSBDeviceById
'IUnattended_getPostInstallScriptTemplatePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getPostInstallScriptTemplatePath
'ISerialPort_getIOBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getIOBase
'IMachine_removeAllCPUIDLeaves' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_removeAllCPUIDLeaves
'IMouseCapabilityChangedEvent_getSupportsAbsolute' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouseCapabilityChangedEvent_getSupportsAbsolute
'IHost_findUSBDeviceById' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_findUSBDeviceById
'IMachine_setIOCacheEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IOCacheEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setIOCacheEnabled
'IUSBDeviceFilters_insertDeviceFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'position', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'filter', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilters_insertDeviceFilter
'INATNetworkPortForwardEvent_getIpv6' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getIpv6
'IVirtualBox_removeCloudNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'network', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_removeCloudNetwork
'IProgress_getPercent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getPercent
'IVirtualBox_getDVDImages' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getDVDImages
'IAudioAdapter_setProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setProperty
'IFsObjInfo_getModificationTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getModificationTime
'INetworkAdapter_setAttachmentType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'attachmentType', type => 'vbox:NetworkAttachmentType', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setAttachmentType
'IHost_findHostNetworkInterfaceById' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_findHostNetworkInterfaceById
'IProcess_writeArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'handle', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:ProcessInputFlag', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IProcess_writeArray
'IMachine_getFirmwareType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getFirmwareType
'INATNetworkSettingEvent_getNeedDhcpServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkSettingEvent_getNeedDhcpServer
'IVirtualBox_getSettingsFilePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getSettingsFilePath
'IDisplay_queryFramebuffer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_queryFramebuffer
'IUnattended_prepare' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_prepare
'IHostPCIDevicePlugEvent_getPlugged' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostPCIDevicePlugEvent_getPlugged
'ICertificate_getSerialNumber' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSerialNumber
'IMachine_getAccessible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAccessible
'IHostNetworkInterface_enableStaticIPConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IPAddress', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkMask', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_enableStaticIPConfig
'IMouseCapabilityChangedEvent_getNeedsHostCursor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouseCapabilityChangedEvent_getNeedsHostCursor
'IGraphicsAdapter_setAccelerate3DEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'accelerate3DEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_setAccelerate3DEnabled
'IHost_removeUSBDeviceSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_removeUSBDeviceSource
'IVirtualBox_getCloudProviderManager' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getCloudProviderManager
'IMachine_setHardwareUUID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hardwareUUID', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setHardwareUUID
'INATRedirectEvent_getProto' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getProto
'IUnattended_setValidationKitIsoPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'validationKitIsoPath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setValidationKitIsoPath
'IGuestPropertyChangedEvent_getFlags' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestPropertyChangedEvent_getFlags
'ISerialPort_getUartType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getUartType
'IRecordingSettings_getScreenSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingSettings_getScreenSettings
'IUSBDeviceFilter_setVendorId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'vendorId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setVendorId
'INetworkAdapter_getSlot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getSlot
'IMachineDebugger_readVirtualMemory' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'address', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_readVirtualMemory
'IRecordingScreenSettings_getOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getOptions
'IGuestFile_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFile_getMidlDoesNotLikeEmptyInterfaces
'IMachine_getParallelPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'slot', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getParallelPort
'IRuntimeErrorEvent_getFatal' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRuntimeErrorEvent_getFatal
'IVirtualBox_getAPIVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getAPIVersion
'IMachine_getCPUCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUCount
'ISystemProperties_getSupportedAutostopTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedAutostopTypes
'IConsole_getMouse' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getMouse
'ICloudClient_importInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'progress', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_importInstance
'IGuestSession_fsObjCopyArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FileCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjCopyArray
'IMouse_getMultiTouchSupported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouse_getMultiTouchSupported
'IGuestDnDSource_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestDnDSource_getMidlDoesNotLikeEmptyInterfaces
'IAppliance_read' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'file', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_read
'ICloudProfile_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_setName
'IProcess_getExecutablePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getExecutablePath
'IExtraDataChangedEvent_getMachineId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IExtraDataChangedEvent_getMachineId
'INATNetwork_getLoopbackIp6' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getLoopbackIp6
'IDHCPGroupCondition_remove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_remove
'IMedium_getVariant' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getVariant
'IGuest_getAdditionsStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'level', type => 'vbox:AdditionsRunLevelType', attr => {}),
    ], # end parameters
  }, # end IGuest_getAdditionsStatus
'IFramebuffer_getHeightReduction' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getHeightReduction
'IEventSource_fireEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'event', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IEventSource_fireEvent
'IGuest_setCredentials' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'userName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'domain', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'allowInteractiveLogon', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuest_setCredentials
'IUSBDeviceFilter_setManufacturer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'manufacturer', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setManufacturer
'IUnattended_setMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'machine', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setMachine
'IDisplay_invalidateAndUpdate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDisplay_invalidateAndUpdate
'IVirtualBox_registerMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'machine', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_registerMachine
'INATEngine_setDNSUseHostResolver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'DNSUseHostResolver', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATEngine_setDNSUseHostResolver
'IRangedIntegerFormValue_setInteger' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IRangedIntegerFormValue_setInteger
'IHost_getProcessorCoreCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorCoreCount
'IMachine_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getId
'IMachineDebugger_getLogEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogEnabled
'ISystemProperties_getSupportedVFSTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedVFSTypes
'IGuestSession_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getStatus
'IMachine_getBIOSSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getBIOSSettings
'IPCIAddress_asLong' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_asLong
'IHost_getUTCTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getUTCTime
'IMachine_findSnapshot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'nameOrId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_findSnapshot
'IMachine_getMediumAttachment' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMachine_getMediumAttachment
'IDirectory_getFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDirectory_getFilter
'IBIOSSettings_getIOAPICEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getIOAPICEnabled
'ISharedFolder_setAutoMount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autoMount', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_setAutoMount
'IUSBDeviceFilter_getSerialNumber' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getSerialNumber
'ISnapshotTakenEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshotTakenEvent_getMidlDoesNotLikeEmptyInterfaces
'IMachine_getDnDMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getDnDMode
'IEventSource_unregisterListener' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'listener', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventSource_unregisterListener
'IGuestProcessIOEvent_getProcessed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessIOEvent_getProcessed
'ICloudClient_listInstances' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'machineState', type => 'vbox:CloudMachineState', attr => {}),
    ], # end parameters
  }, # end ICloudClient_listInstances
'IRecordingScreenSettings_getMaxTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getMaxTime
'IMachineDebugger_getPAEEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getPAEEnabled
'ISystemProperties_getStorageControllerHotplugCapable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerType', type => 'vbox:StorageControllerType', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getStorageControllerHotplugCapable
'IMachine_getSettingsModified' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSettingsModified
'IVirtualBox_getCloudNetworks' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getCloudNetworks
'ICertificate_getPublicKeyAlgorithmOID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getPublicKeyAlgorithmOID
'IGuestSession_directoryCopyToGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:DirectoryCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryCopyToGuest
'IMachine_getHardwareVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getHardwareVersion
'ISnapshot_getChildren' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getChildren
'IBIOSSettings_getLogoImagePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getLogoImagePath
'IBandwidthControl_createBandwidthGroup' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:BandwidthGroupType', attr => {}),
      SOAP::Data->new(name => 'maxBytesPerSec', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IBandwidthControl_createBandwidthGroup
'IProcess_getEnvironment' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getEnvironment
'INATRedirectEvent_getSlot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getSlot
'IUnattended_setInstallGuestAdditions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'installGuestAdditions', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IUnattended_setInstallGuestAdditions
'ISnapshot_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_setName
'IFsObjInfo_getAllocatedSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getAllocatedSize
'ISystemProperties_getSupportedRecordingAudioCodecs' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedRecordingAudioCodecs
'INATEngine_setAliasMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'aliasMode', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end INATEngine_setAliasMode
'IGraphicsAdapter_getAccelerate2DVideoEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_getAccelerate2DVideoEnabled
'ICloudClient_setupCloudNetworkEnvironment' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'tunnelNetworkName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'tunnelNetworkRange', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'gatewayOsName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'gatewayOsVersion', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'gatewayShape', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_setupCloudNetworkEnvironment
'IMachine_getSessionState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSessionState
'IUnattended_setPackageSelectionAdjustments' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'packageSelectionAdjustments', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setPackageSelectionAdjustments
'ISystemProperties_getStorageBusForStorageControllerType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'storageControllerType', type => 'vbox:StorageControllerType', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getStorageBusForStorageControllerType
'IMachine_setSnapshotFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'snapshotFolder', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setSnapshotFolder
'IAppliance_getMediumIdsForPasswordId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'passwordId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getMediumIdsForPasswordId
'IVirtualBox_getVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getVersion
'IUSBDeviceFilter_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setName
'IConsole_findUSBDeviceByAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_findUSBDeviceByAddress
'IMachine_getClipboardFileTransfersEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getClipboardFileTransfersEnabled
'IMachine_getTeleporterPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getTeleporterPassword
'IUnattended_getProductKey' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getProductKey
'IVirtualBox_getMachinesByGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'groups', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getMachinesByGroups
'IDirectory_read' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDirectory_read
'IConsole_getKeyboard' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getKeyboard
'IStorageController_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getName
'IConsole_teleport' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostname', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'tcpport', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maxDowntime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IConsole_teleport
'IVirtualBox_removeSharedFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_removeSharedFolder
'IDisplay_notifyHiDPIOutputPolicyChange' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'fUnscaledHiDPI', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IDisplay_notifyHiDPIOutputPolicyChange
'ISession_getConsole' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_getConsole
'IHost_getVideoInputDevices' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getVideoInputDevices
'IDHCPConfig_getMaxLeaseTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getMaxLeaseTime
'IUSBDevice_getVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getVersion
'ISharedFolder_getAccessible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getAccessible
'IVRDEServer_getAuthTimeout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getAuthTimeout
'INATRedirectEvent_getGuestPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getGuestPort
'IProgress_getErrorInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getErrorInfo
'IGuestMonitorChangedEvent_getChangeType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorChangedEvent_getChangeType
'IEvent_waitProcessed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IEvent_waitProcessed
'INATEngine_setHostIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostIP', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_setHostIP
'IFramebuffer_getPixelFormat' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getPixelFormat
'IGuestMultiTouchEvent_getContactFlags' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMultiTouchEvent_getContactFlags
'IMachineDebugger_info' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'args', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_info
'IVirtualBox_getExtraDataKeys' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getExtraDataKeys
'INetworkAdapter_setCloudNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cloudNetwork', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setCloudNetwork
'INATNetworkPortForwardEvent_getProto' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getProto
'IHost_getDVDDrives' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getDVDDrives
'IMedium_lockWrite' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_lockWrite
'IMachine_getRecordingSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getRecordingSettings
'IPerformanceMetric_getMaximumValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getMaximumValue
'IDHCPIndividualConfig_getMACAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPIndividualConfig_getMACAddress
'IProcess_terminate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_terminate
'IRecordingScreenSettings_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getEnabled
'IMedium_reset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_reset
'IRangedIntegerFormValue_getSuffix' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRangedIntegerFormValue_getSuffix
'IMachine_detachDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMachine_detachDevice
'IAdditionsStateChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAdditionsStateChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'IMachineDebugger_setPATMEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'PATMEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setPATMEnabled
'IHost_generateMACAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_generateMACAddress
'IProcess_getPID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getPID
'ICloudNetworkGatewayInfo_getSecondaryPublicIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetworkGatewayInfo_getSecondaryPublicIP
'IStorageController_getMaxDevicesPerPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getMaxDevicesPerPortCount
'ISystemProperties_getSupportedPointingHIDTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedPointingHIDTypes
'IMachine_getTeleporterEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getTeleporterEnabled
'IVetoEvent_isApproved' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVetoEvent_isApproved
'IGuestSession_fileCreateTemp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'templateName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'secure', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileCreateTemp
'IHostNetworkInterface_getNetworkMask' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getNetworkMask
'IRecordingScreenSettings_setMaxFileSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maxFileSize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setMaxFileSize
'IMediumChangedEvent_getMediumAttachment' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumChangedEvent_getMediumAttachment
'IGuestFileIOEvent_getProcessed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileIOEvent_getProcessed
'IDisplay_invalidateAndUpdateScreen' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_invalidateAndUpdateScreen
'ICloudClient_listImages' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'imageState', type => 'vbox:CloudImageState', attr => {}),
    ], # end parameters
  }, # end ICloudClient_listImages
'IUnattended_setPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setPassword
'IMachine_getMemorySize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getMemorySize
'IVRDEServer_setVRDEProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setVRDEProperty
'IAudioAdapter_getAudioCodec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getAudioCodec
'IMachine_deleteSnapshot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_deleteSnapshot
'IVirtualBox_getFloppyImages' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getFloppyImages
'INATNetwork_setAdvertiseDefaultIPv6RouteEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'advertiseDefaultIPv6RouteEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setAdvertiseDefaultIPv6RouteEnabled
'IGuestSession_fileOpenEx' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'accessMode', type => 'vbox:FileAccessMode', attr => {}),
      SOAP::Data->new(name => 'openAction', type => 'vbox:FileOpenAction', attr => {}),
      SOAP::Data->new(name => 'sharingMode', type => 'vbox:FileSharingMode', attr => {}),
      SOAP::Data->new(name => 'creationMode', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FileOpenExFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileOpenEx
'IUSBProxyBackend_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBProxyBackend_getType
'IMedium_resize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logicalSize', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IMedium_resize
'IGuestProcessEvent_getProcess' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessEvent_getProcess
'IMachineDebugger_getRegisters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getRegisters
'IMachine_setClipboardMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'clipboardMode', type => 'vbox:ClipboardMode', attr => {}),
    ], # end parameters
  }, # end IMachine_setClipboardMode
'IMediumIO_read' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMediumIO_read
'IVirtualBox_getVersionNormalized' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getVersionNormalized
'INetworkAdapter_getProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getProperties
'IRecordingScreenSettings_getVideoHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoHeight
'INATNetworkSettingEvent_getNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkSettingEvent_getNetwork
'IUSBDevice_getSpeed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getSpeed
'IFsObjInfo_getObjectSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getObjectSize
'IMachineDebugger_dumpHostProcessCore' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'filename', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'compression', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_dumpHostProcessCore
'IGuestSession_setCurrentDirectory' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'currentDirectory', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_setCurrentDirectory
'IRecordingScreenSettings_setAudioHz' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioHz', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setAudioHz
'IMachine_setCPUCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'CPUCount', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUCount
'IBIOSSettings_getLogoFadeIn' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getLogoFadeIn
'ISession_getMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_getMachine
'IDnDTarget_leave' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDnDTarget_leave
'IFramebufferOverlay_setAlpha' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'alpha', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_setAlpha
'IMachine_attachHostPCIDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostAddress', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'desiredGuestAddress', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'tryToUnbind', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_attachHostPCIDevice
'IRecordingScreenSettings_setMaxTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maxTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setMaxTime
'IProgress_getOperationPercent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getOperationPercent
'IHostPCIDevicePlugEvent_getMessage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostPCIDevicePlugEvent_getMessage
'IMachineDebugger_getExecuteAllInIEM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getExecuteAllInIEM
'IMachine_getHardwareUUID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getHardwareUUID
'INATEngine_getNetworkSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getNetworkSettings
'IGuestSession_getEnvironmentBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getEnvironmentBase
'ICloudProfile_setProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'values', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_setProperties
'IConsole_getAttachedPCIDevices' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getAttachedPCIDevices
'IMediumRegisteredEvent_getRegistered' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumRegisteredEvent_getRegistered
'IMachine_discardSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_discardSettings
'IGuest_getDnDSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getDnDSource
'INATEngine_getAliasMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getAliasMode
'IMachine_setFirmwareType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'firmwareType', type => 'vbox:FirmwareType', attr => {}),
    ], # end parameters
  }, # end IMachine_setFirmwareType
'ICloudClient_getInstanceInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_getInstanceInfo
'IMedium_setLocation' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'location', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_setLocation
'INetworkAdapter_setGenericDriver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'genericDriver', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setGenericDriver
'IVirtualBox_findDHCPServerByNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_findDHCPServerByNetworkName
'ISerialPort_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setEnabled
'IStorageController_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_setName
'IMachine_setHardwareVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hardwareVersion', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setHardwareVersion
'IUnattended_setPostInstallScriptTemplatePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'postInstallScriptTemplatePath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setPostInstallScriptTemplatePath
'IChoiceFormValue_getSelectedIndex' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IChoiceFormValue_getSelectedIndex
'IGuestFsObjInfo_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFsObjInfo_getMidlDoesNotLikeEmptyInterfaces
'IDHCPGroupCondition_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_getType
'IVirtualBox_getExtraData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getExtraData
'INetworkAdapter_setBandwidthGroup' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bandwidthGroup', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setBandwidthGroup
'IVirtualBox_getSharedFolders' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getSharedFolders
'IMouse_putMouseEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'dx', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'dy', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'dz', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'dw', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'buttonState', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMouse_putMouseEvent
'IFramebufferOverlay_move' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFramebufferOverlay_move
'IUSBDeviceFilters_getDeviceFilters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilters_getDeviceFilters
'IVRDEServerInfoChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServerInfoChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'IMachineDebugger_writePhysicalMemory' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'address', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bytes', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_writePhysicalMemory
'IDHCPIndividualConfig_getFixedAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPIndividualConfig_getFixedAddress
'ISnapshot_setDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_setDescription
'IBooleanFormValue_setSelected' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'selected', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBooleanFormValue_setSelected
'ICloudClient_createImage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'parameters', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_createImage
'IConsole_createSharedFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostPath', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'writable', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'automount', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'autoMountPoint', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_createSharedFolder
'IRecordingChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'IVetoEvent_addVeto' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'reason', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVetoEvent_addVeto
'IVirtualBox_getPerformanceCollector' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getPerformanceCollector
'IVBoxSVCAvailabilityChangedEvent_getAvailable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVBoxSVCAvailabilityChangedEvent_getAvailable
'IConsole_getDeviceActivity' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:DeviceType', attr => {}),
    ], # end parameters
  }, # end IConsole_getDeviceActivity
'IConsole_getGuestEnteredACPIMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getGuestEnteredACPIMode
'ISystemProperties_getMinGuestVRAM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMinGuestVRAM
'IVetoEvent_addApproval' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'reason', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVetoEvent_addApproval
'IUnattended_setTimeZone' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeZone', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setTimeZone
'IDHCPConfig_getSuppressedOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getSuppressedOptions
'IConsole_attachUSBDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'captureFilename', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_attachUSBDevice
'ISystemProperties_getMaxGuestCPUCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxGuestCPUCount
'IProcess_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getStatus
'INetworkAdapter_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getEnabled
'IMachine_getMediumAttachments' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getMediumAttachments
'IRuntimeErrorEvent_getMessage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRuntimeErrorEvent_getMessage
'IManagedObjectRef_getInterfaceName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IManagedObjectRef_getInterfaceName
'IConsole_removeDiskEncryptionPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_removeDiskEncryptionPassword
'IBandwidthGroup_getMaxBytesPerSec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthGroup_getMaxBytesPerSec
'IRecordingScreenSettings_setOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'options', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setOptions
'ICloudProvider_restoreProfiles' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_restoreProfiles
'IRecordingScreenSettings_getAudioHz' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getAudioHz
'IFile_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getId
'IMachine_setDnDMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'dnDMode', type => 'vbox:DnDMode', attr => {}),
    ], # end parameters
  }, # end IMachine_setDnDMode
'IProgress_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getEventSource
'IRecordingScreenSettings_getVideoScalingMethod' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoScalingMethod
'IAudioAdapter_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setEnabled
'IUnattended_setExtraInstallKernelParameters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'extraInstallKernelParameters', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setExtraInstallKernelParameters
'IBandwidthGroup_getReference' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthGroup_getReference
'IUSBDeviceFilter_setProduct' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'product', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setProduct
'ISystemProperties_getSupportedKeyboardHIDTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedKeyboardHIDTypes
'INATNetwork_start' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_start
'IUnattended_getScriptTemplatePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getScriptTemplatePath
'IUSBDeviceFilter_getPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getPort
'IParallelPort_setPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPort_setPath
'IMachine_discardSavedState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'fRemoveFile', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_discardSavedState
'IGuestUserStateChangedEvent_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestUserStateChangedEvent_getName
'IVirtualBoxErrorInfo_getResultCode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBoxErrorInfo_getResultCode
'IMedium_lockRead' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_lockRead
'IFsObjInfo_getNodeIdDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getNodeIdDevice
'IDHCPIndividualConfig_getSlot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPIndividualConfig_getSlot
'IVirtualBox_openMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'location', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'deviceType', type => 'vbox:DeviceType', attr => {}),
      SOAP::Data->new(name => 'accessMode', type => 'vbox:AccessMode', attr => {}),
      SOAP::Data->new(name => 'forceNewUuid', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_openMedium
'IMedium_getAutoReset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getAutoReset
'IAppliance_importMachines' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ImportOptions', attr => {}),
    ], # end parameters
  }, # end IAppliance_importMachines
'IUSBDeviceFilter_setMaskedInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maskedInterfaces', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setMaskedInterfaces
'INATNetwork_stop' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_stop
'IUSBDevice_getDeviceInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getDeviceInfo
'IMouseCapabilityChangedEvent_getSupportsRelative' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouseCapabilityChangedEvent_getSupportsRelative
'IEventSource_registerListener' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'listener', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'interesting', type => 'vbox:VBoxEventType', attr => {}),
      SOAP::Data->new(name => 'active', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IEventSource_registerListener
'IMachine_setAutoDiscardForDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'discard', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setAutoDiscardForDevice
'ISystemProperties_getSupportedStorageControllerTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedStorageControllerTypes
'ICertificate_getValidityPeriodNotBefore' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getValidityPeriodNotBefore
'ISystemProperties_getSupportedCloneOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedCloneOptions
'IRuntimeErrorEvent_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRuntimeErrorEvent_getId
'IFramebuffer_notifyUpdateImage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'image', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_notifyUpdateImage
'IHostNetworkInterface_getIPV6NetworkMaskPrefixLength' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getIPV6NetworkMaskPrefixLength
'IForm_getValues' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IForm_getValues
'IDHCPConfig_getAllOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getAllOptions
'ISystemProperties_setProxyMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'proxyMode', type => 'vbox:ProxyMode', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setProxyMode
'INATNetworkSettingEvent_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkSettingEvent_getEnabled
'IHostNetworkInterface_getInterfaceType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getInterfaceType
'IFile_setACL' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'acl', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFile_setACL
'IGuestSession_getEnvironmentChanges' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getEnvironmentChanges
'IWebsessionManager_logoff' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => 'refIVirtualBox', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IWebsessionManager_logoff
'IVirtualSystemDescription_getDescriptionByType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:VirtualSystemDescriptionType', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_getDescriptionByType
'IUnattended_getHostname' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getHostname
'ISystemProperties_getProxyURL' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getProxyURL
'IMachine_getMemoryBalloonSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getMemoryBalloonSize
'IGuest_getMemoryBalloonSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getMemoryBalloonSize
'ISystemProperties_getMaxNetworkAdaptersOfType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'chipset', type => 'vbox:ChipsetType', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:NetworkAttachmentType', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxNetworkAdaptersOfType
'IMachine_setTeleporterEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'teleporterEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setTeleporterEnabled
'IMachine_setPointingHIDType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pointingHIDType', type => 'vbox:PointingHIDType', attr => {}),
    ], # end parameters
  }, # end IMachine_setPointingHIDType
'IRecordingScreenSettings_getAudioChannels' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getAudioChannels
'IMachine_getCPUIDPortabilityLevel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUIDPortabilityLevel
'IVirtualBox_createNATNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createNATNetwork
'INATEngine_removeRedirect' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_removeRedirect
'IFsObjInfo_getNodeId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getNodeId
'IMachine_getCPUIDLeaf' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'idx', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'idxSub', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUIDLeaf
'ISystemProperties_getAutostartDatabasePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getAutostartDatabasePath
'IHost_getFloppyDrives' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getFloppyDrives
'IMachine_getAutostartEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAutostartEnabled
'IMachine_getSessionName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSessionName
'IMachineDebugger_modifyLogDestinations' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settings', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_modifyLogDestinations
'IMachine_getTeleporterPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getTeleporterPort
'IDisplay_getVideoModeHint' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'display', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_getVideoModeHint
'IDisplay_takeScreenShotToArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bitmapFormat', type => 'vbox:BitmapFormat', attr => {}),
    ], # end parameters
  }, # end IDisplay_takeScreenShotToArray
'IGuest_getAdditionsVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getAdditionsVersion
'IMachine_setDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setDescription
'ISharedFolderChangedEvent_getScope' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolderChangedEvent_getScope
'IFile_queryInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_queryInfo
'ISystemProperties_getDefaultIoCacheSettingForStorageController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerType', type => 'vbox:StorageControllerType', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultIoCacheSettingForStorageController
'IConsole_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getState
'IMachine_setKeyboardHIDType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'keyboardHIDType', type => 'vbox:KeyboardHIDType', attr => {}),
    ], # end parameters
  }, # end IMachine_setKeyboardHIDType
'ICloudProfile_remove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_remove
'IGuestMouseEvent_getX' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMouseEvent_getX
'IBandwidthGroup_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthGroup_getType
'IGuestSession_fileCopyToGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FileCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileCopyToGuest
'IParallelPort_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPort_getEnabled
'IProgress_waitForOperationCompletion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'operation', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IProgress_waitForOperationCompletion
'IToken_abandon' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IToken_abandon
'IDisplay_setSeamlessMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IDisplay_setSeamlessMode
'IStorageControllerChangedEvent_getControllerName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageControllerChangedEvent_getControllerName
'INATNetwork_getPortForwardRules4' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getPortForwardRules4
'IHostVideoInputDevice_getAlias' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostVideoInputDevice_getAlias
'IMachine_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setName
'IDHCPServer_getGlobalConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getGlobalConfig
'INATNetworkSettingEvent_getGateway' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkSettingEvent_getGateway
'ISystemProperties_getParallelPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getParallelPortCount
'IAppliance_interpret' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_interpret
'IMachine_getGraphicsAdapter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getGraphicsAdapter
'ICloudNetwork_getProvider' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_getProvider
'IKeyboard_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboard_getEventSource
'IDHCPConfig_getMinLeaseTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getMinLeaseTime
'IMachine_setCPUExecutionCap' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'CPUExecutionCap', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUExecutionCap
'IBIOSSettings_getBootMenuMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getBootMenuMode
'IUnattended_setImageIndex' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'imageIndex', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IUnattended_setImageIndex
'IDirectory_close' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDirectory_close
'INetworkAdapter_setTraceEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'traceEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setTraceEnabled
'ISystemProperties_getSerialPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSerialPortCount
'IMachine_setTracingEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'tracingEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setTracingEnabled
'IDHCPConfig_getScope' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getScope
'INetworkAdapter_setMACAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'MACAddress', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setMACAddress
'IConsole_sleepButton' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_sleepButton
'IMachine_getStorageControllers' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getStorageControllers
'IMachineDebugger_setVirtualTimeRate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'virtualTimeRate', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setVirtualTimeRate
'IMedium_mergeTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_mergeTo
'IGuestUserStateChangedEvent_getStateDetails' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestUserStateChangedEvent_getStateDetails
'IFormValue_getHelp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getHelp
'IHostUSBDeviceFilter_setAction' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'action', type => 'vbox:USBDeviceFilterAction', attr => {}),
    ], # end parameters
  }, # end IHostUSBDeviceFilter_setAction
'INATEngine_getTFTPPrefix' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getTFTPPrefix
'IMachine_saveState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_saveState
'IMachine_saveSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_saveSettings
'IMachine_setNoBandwidthGroupForDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMachine_setNoBandwidthGroupForDevice
'IConsole_powerButton' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_powerButton
'IEventSource_createAggregator' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'subordinates', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventSource_createAggregator
'IGuestSessionStateChangedEvent_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSessionStateChangedEvent_getStatus
'ICloudProfile_createCloudClient' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_createCloudClient
'IDisplay_attachFramebuffer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'framebuffer', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDisplay_attachFramebuffer
'INetworkAdapter_getInternalNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getInternalNetwork
'IMachine_getNetworkAdapter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'slot', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getNetworkAdapter
'IMachine_getUSBControllers' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getUSBControllers
'ISystemProperties_getSupportedImportOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedImportOptions
'IMedium_createBaseStorage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logicalSize', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'variant', type => 'vbox:MediumVariant', attr => {}),
    ], # end parameters
  }, # end IMedium_createBaseStorage
'IVirtualBox_createMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settingsFile', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'groups', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'osTypeId', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createMachine
'IEventSourceChangedEvent_getListener' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventSourceChangedEvent_getListener
'ICloudClient_getLaunchDescriptionForm' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_getLaunchDescriptionForm
'IMachineDebugger_getOSVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getOSVersion
'IHost_findHostNetworkInterfaceByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_findHostNetworkInterfaceByName
'IMachine_setEmulatedUSBCardReaderEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'emulatedUSBCardReaderEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setEmulatedUSBCardReaderEnabled
'IEvent_setProcessed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEvent_setProcessed
'IMachine_getCPUIDLeafByOrdinal' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'ordinal', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUIDLeafByOrdinal
'IHost_getDomainName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getDomainName
'IBooleanFormValue_getSelected' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBooleanFormValue_getSelected
'IUnattended_done' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_done
'IFile_getOpenAction' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getOpenAction
'IManagedObjectRef_release' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IManagedObjectRef_release
'IMachine_cloneTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'target', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'vbox:CloneMode', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:CloneOptions', attr => {}),
    ], # end parameters
  }, # end IMachine_cloneTo
'IVirtualBox_findCloudNetworkByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_findCloudNetworkByName
'ICloudClient_getExportDescriptionForm' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'description', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_getExportDescriptionForm
'IPerformanceMetric_getUnit' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getUnit
'ISystemProperties_getSupportedAudioDriverTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedAudioDriverTypes
'IAudioAdapter_getAudioController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getAudioController
'IGuestSession_fsObjMoveArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FsObjMoveFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjMoveArray
'IAudioAdapter_setAudioController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioController', type => 'vbox:AudioControllerType', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setAudioController
'IMachineDebugger_getHWVirtExVPIDEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getHWVirtExVPIDEnabled
'IBIOSSettings_getSMBIOSUuidLittleEndian' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getSMBIOSUuidLittleEndian
'ICloudNetworkEnvironmentInfo_getTunnelNetworkId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetworkEnvironmentInfo_getTunnelNetworkId
'IProgress_waitForCompletion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IProgress_waitForCompletion
'IKeyboardLedsChangedEvent_getCapsLock' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboardLedsChangedEvent_getCapsLock
'IUSBDeviceFilter_setSerialNumber' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'serialNumber', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setSerialNumber
'IMachine_getGuestProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getGuestProperty
'IBIOSSettings_setTimeOffset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeOffset', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setTimeOffset
'IFile_write' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFile_write
'IDHCPServer_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getEventSource
'IMachine_setBandwidthGroupForDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'bandwidthGroup', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setBandwidthGroupForDevice
'IGuestSessionStateChangedEvent_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSessionStateChangedEvent_getId
'IGuestSession_fsObjRename' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'oldPath', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'newPath', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FsObjRenameFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjRename
'IRangedIntegerFormValue_getMaximum' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRangedIntegerFormValue_getMaximum
'IMachine_getTeleporterAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getTeleporterAddress
'IMediumIO_write' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumIO_write
'ICertificate_getExtendedKeyUsage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getExtendedKeyUsage
'IParallelPort_getSlot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPort_getSlot
'IGuestSession_fileOpen' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'accessMode', type => 'vbox:FileAccessMode', attr => {}),
      SOAP::Data->new(name => 'openAction', type => 'vbox:FileOpenAction', attr => {}),
      SOAP::Data->new(name => 'creationMode', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileOpen
'INATEngine_getRedirects' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getRedirects
'IMousePointerShapeChangedEvent_getXhot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getXhot
'ISystemProperties_getFreeDiskSpaceError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getFreeDiskSpaceError
'INetworkAdapter_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setEnabled
'IHost_getUpdate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getUpdate
'ICloudProvider_getProfileNames' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getProfileNames
'IGuest_updateGuestAdditions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'arguments', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:AdditionsUpdateFlag', attr => {}),
    ], # end parameters
  }, # end IGuest_updateGuestAdditions
'INATNetwork_getAdvertiseDefaultIPv6RouteEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getAdvertiseDefaultIPv6RouteEnabled
'ISnapshot_getChildrenCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getChildrenCount
'IGuestProcessIOEvent_getHandle' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessIOEvent_getHandle
'IVirtualBox_getHost' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getHost
'IGuestSession_directoryCopyFromGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:DirectoryCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryCopyFromGuest
'INetworkAdapter_getCloudNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getCloudNetwork
'INATNetworkPortForwardEvent_getGuestPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getGuestPort
'IMedium_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getState
'IConsole_powerUpPaused' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_powerUpPaused
'IUSBDeviceStateChangedEvent_getAttached' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceStateChangedEvent_getAttached
'IVRDEServer_setVRDEExtPack' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'VRDEExtPack', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setVRDEExtPack
'IVRDEServer_getAuthLibrary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getAuthLibrary
'ISnapshotDeletedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshotDeletedEvent_getMidlDoesNotLikeEmptyInterfaces
'IGuest_getOSTypeId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getOSTypeId
'INATNetworkCreationDeletionEvent_getCreationEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkCreationDeletionEvent_getCreationEvent
'IVirtualBox_createAppliance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createAppliance
'IDisplay_getScreenResolution' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_getScreenResolution
'IDHCPGroupCondition_setValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_setValue
'INetworkAdapter_setPromiscModePolicy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'promiscModePolicy', type => 'vbox:NetworkAdapterPromiscModePolicy', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setPromiscModePolicy
'IParallelPort_setIOBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IOBase', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IParallelPort_setIOBase
'INATRedirectEvent_getRemove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getRemove
'ISharedFolder_getLastAccessError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getLastAccessError
'INATEngine_getDNSUseHostResolver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getDNSUseHostResolver
'IDisplay_notifyScaleFactorChange' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'u32ScaleFactorWMultiplied', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'u32ScaleFactorHMultiplied', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDisplay_notifyScaleFactorChange
'ISerialPortChangedEvent_getSerialPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPortChangedEvent_getSerialPort
'IVirtualBox_getHardDisks' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getHardDisks
'ISystemProperties_setFreeDiskSpaceWarning' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'freeDiskSpaceWarning', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setFreeDiskSpaceWarning
'IFile_getInitialSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getInitialSize
'IGuestDirectory_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestDirectory_getMidlDoesNotLikeEmptyInterfaces
'IMedium_getSnapshotIds' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'machineId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getSnapshotIds
'IUSBProxyBackend_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBProxyBackend_getName
'IGuestSession_environmentScheduleSet' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_environmentScheduleSet
'IVirtualBoxErrorInfo_getInterfaceID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBoxErrorInfo_getInterfaceID
'IMachineDebugger_setRegister' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setRegister
'INetworkAdapter_getTraceFile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getTraceFile
'ICloudProvider_getProfiles' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getProfiles
'IRecordingScreenSettings_getVideoWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoWidth
'IUSBDevice_getAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getAddress
'ICloudClient_pauseInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_pauseInstance
'IUSBDevice_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getId
'IMachine_hotUnplugCPU' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpu', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_hotUnplugCPU
'IShowWindowEvent_getWinId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IShowWindowEvent_getWinId
'IGuest_findSession' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'sessionName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_findSession
'IMediumIO_close' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumIO_close
'IGuest_getSessions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getSessions
'IGuestSession_getProcesses' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getProcesses
'ISnapshot_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getDescription
'IParallelPort_getIRQ' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPort_getIRQ
'IUSBDeviceFilter_getRevision' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getRevision
'IHostNetworkInterface_getShortName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getShortName
'ISnapshotEvent_getSnapshotId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshotEvent_getSnapshotId
'IMediumIO_getExplorer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumIO_getExplorer
'ICloudNetwork_setNetworkId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_setNetworkId
'IGuestScreenInfo_getOriginX' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getOriginX
'IKeyboard_getKeyboardLEDs' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboard_getKeyboardLEDs
'IMachine_addStorageController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'connectionType', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end IMachine_addStorageController
'IMachine_setBootOrder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'position', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'device', type => 'vbox:DeviceType', attr => {}),
    ], # end parameters
  }, # end IMachine_setBootOrder
'IDHCPServer_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_setEnabled
'IPCIAddress_setDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:short', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_setDevice
'IConsole_getVRDEServerInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getVRDEServerInfo
'ISerialPort_setPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setPath
'ISession_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_getName
'IMachine_setIOCacheSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IOCacheSize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setIOCacheSize
'IMachine_setPageFusionEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pageFusionEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setPageFusionEnabled
'ICloudProvider_saveProfiles' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_saveProfiles
'IGraphicsAdapter_setVRAMSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'VRAMSize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_setVRAMSize
'IUnattended_setIsoPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'isoPath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setIsoPath
'IDHCPServer_setConfiguration' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IPAddress', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkMask', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'FromIPAddress', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'ToIPAddress', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_setConfiguration
'IHostNetworkInterface_getStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getStatus
'IUSBDevice_getProduct' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getProduct
'IVRDEServer_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getEnabled
'INATNetwork_addLocalMapping' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostid', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end INATNetwork_addLocalMapping
'ISystemProperties_getMaxGuestVRAM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxGuestVRAM
'IMachine_canShowConsoleWindow' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_canShowConsoleWindow
'IMedium_getSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getSize
'INetworkAdapter_setAdapterType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'adapterType', type => 'vbox:NetworkAdapterType', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setAdapterType
'IMousePointerShape_getVisible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getVisible
'IRecordingScreenSettings_setAudioCodec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioCodec', type => 'vbox:RecordingAudioCodec', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setAudioCodec
'IAudioAdapter_getPropertiesList' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getPropertiesList
'IVirtualSystemDescriptionForm_getVirtualSystemDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescriptionForm_getVirtualSystemDescription
'IUSBDeviceFilter_setRemote' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'remote', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setRemote
'ICloudClient_exportImage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'image', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'parameters', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_exportImage
'ICursorPositionChangedEvent_getHasData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICursorPositionChangedEvent_getHasData
'IMachine_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getName
'IFormValue_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getType
'ISystemProperties_getMaxPortCountForStorageBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bus', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxPortCountForStorageBus
'INetworkAdapter_getGenericDriver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getGenericDriver
'INATEngine_setTFTPPrefix' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'TFTPPrefix', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_setTFTPPrefix
'INATRedirectEvent_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getName
'ICloudProviderManager_getProviderByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'providerName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProviderManager_getProviderByName
'IUnattended_getInstallTestExecService' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getInstallTestExecService
'INATNetwork_setNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setNetworkName
'IProgress_cancel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_cancel
'IUnattended_getProxy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getProxy
'IMachine_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getDescription
'INATRedirectEvent_getHostPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getHostPort
'INATEngine_setTFTPBootFile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'TFTPBootFile', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_setTFTPBootFile
'ISystemProperties_setExclusiveHwVirt' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'exclusiveHwVirt', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setExclusiveHwVirt
'IMachine_setHWVirtExProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'vbox:HWVirtExPropertyType', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setHWVirtExProperty
'IHost_getOSVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getOSVersion
'ISystemProperties_setDefaultFrontend' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultFrontend', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setDefaultFrontend
'IVirtualBox_getGuestOSType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getGuestOSType
'IStringArray_getValues' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStringArray_getValues
'IVirtualBox_getSystemProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getSystemProperties
'IMachine_setMemoryBalloonSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'memoryBalloonSize', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setMemoryBalloonSize
'IGuestMultiTouchEvent_getContactCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMultiTouchEvent_getContactCount
'IMachine_setCPUIDLeaf' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'idx', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'idxSub', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'valEax', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'valEbx', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'valEcx', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'valEdx', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUIDLeaf
'ISharedFolder_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getName
'ICertificate_getValidityPeriodNotAfter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getValidityPeriodNotAfter
'IVetoEvent_isVetoed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVetoEvent_isVetoed
'IHostUSBDevice_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostUSBDevice_getState
'INATEngine_addRedirect' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'proto', type => 'vbox:NATProtocol', attr => {}),
      SOAP::Data->new(name => 'hostIP', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostPort', type => 'xsd:unsignedShort', attr => {}),
      SOAP::Data->new(name => 'guestIP', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'guestPort', type => 'xsd:unsignedShort', attr => {}),
    ], # end parameters
  }, # end INATEngine_addRedirect
'ICertificate_getSubjectName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSubjectName
'IMachineDebugger_injectNMI' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_injectNMI
'IDnDSource_drop' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'action', type => 'vbox:DnDAction', attr => {}),
    ], # end parameters
  }, # end IDnDSource_drop
'IDnDBase_getFormats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDBase_getFormats
'IMachineDebugger_getHWVirtExUXEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getHWVirtExUXEnabled
'IFile_getCreationMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getCreationMode
'IExtraDataCanChangeEvent_getMachineId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IExtraDataCanChangeEvent_getMachineId
'IConsole_reset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_reset
'IMachine_getUSBControllerCountByType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:USBControllerType', attr => {}),
    ], # end parameters
  }, # end IMachine_getUSBControllerCountByType
'IHost_getSearchStrings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getSearchStrings
'IMachine_passthroughDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'passthrough', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_passthroughDevice
'IMachine_restoreSnapshot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'snapshot', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_restoreSnapshot
'ISystemProperties_setDefaultAdditionsISO' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultAdditionsISO', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setDefaultAdditionsISO
'IFile_getAccessMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getAccessMode
'IUSBDeviceFilter_getActive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getActive
'IVirtualBox_removeNATNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'network', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_removeNATNetwork
'IMachine_getVMProcessPriority' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getVMProcessPriority
'IFile_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getEventSource
'IHostVideoInputDevice_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostVideoInputDevice_getName
'IRecordingScreenSettings_setDestination' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'vbox:RecordingDestination', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setDestination
'IDHCPIndividualConfig_setFixedAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'fixedAddress', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPIndividualConfig_setFixedAddress
'IFsObjInfo_getGID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getGID
'IUnattended_getAuxiliaryBasePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getAuxiliaryBasePath
'IProgress_getTimeout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getTimeout
'IMachineDebugger_detectOS' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_detectOS
'IGuestFileEvent_getFile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileEvent_getFile
'IGraphicsAdapter_getMonitorCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_getMonitorCount
'IGuestSession_getTimeout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getTimeout
'IGuestSession_symlinkRead' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'symlink', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:SymlinkReadFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_symlinkRead
'IUSBDeviceFilter_getVendorId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getVendorId
'IMachine_getSettingsFilePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSettingsFilePath
'IDHCPServer_start' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'trunkName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'trunkType', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_start
'IDnDTarget_cancel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDTarget_cancel
'IHostNetworkInterface_getIPV6Address' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getIPV6Address
'IFsObjInfo_getHardLinks' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getHardLinks
'IGuestSession_getUserDocuments' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getUserDocuments
'ISystemProperties_getDefaultFrontend' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultFrontend
'IMachine_getAllowTracingToAccessVM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAllowTracingToAccessVM
'IGraphicsAdapter_getVRAMSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGraphicsAdapter_getVRAMSize
'IBIOSSettings_setLogoImagePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logoImagePath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setLogoImagePath
'ISystemProperties_getSupportedParavirtProviders' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedParavirtProviders
'ISerialPort_getServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getServer
'IRecordingSettings_getScreens' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingSettings_getScreens
'ICloudProfile_getProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_getProperties
'IGuestSession_fsObjMove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FsObjMoveFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjMove
'IRecordingScreenSettings_setVideoHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoHeight', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoHeight
'IDHCPServer_stop' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_stop
'IVRDEServer_getVRDEProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getVRDEProperties
'IHost_getUSBDeviceFilters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getUSBDeviceFilters
'IMachine_launchVMProcess' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'session', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'environmentChanges', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_launchVMProcess
'IFile_querySize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_querySize
'INATNetwork_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getEventSource
'IMachineDebugger_getLogRelDestinations' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogRelDestinations
'IVRDEServer_setReuseSingleConnection' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'reuseSingleConnection', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setReuseSingleConnection
'IFramebuffer_notify3DEvent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_notify3DEvent
'IPCIAddress_fromLong' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'number', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_fromLong
'IBIOSSettings_setSMBIOSUuidLittleEndian' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'SMBIOSUuidLittleEndian', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setSMBIOSUuidLittleEndian
'ISystemProperties_setVRDEAuthLibrary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'VRDEAuthLibrary', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setVRDEAuthLibrary
'IDisplay_detachScreens' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenIds', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IDisplay_detachScreens
'INATNetwork_setIPv6Enabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IPv6Enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setIPv6Enabled
'ICloudNetwork_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_setEnabled
'IMachineDebugger_getRecompileUser' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getRecompileUser
'IVirtualBox_getNATNetworks' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getNATNetworks
'IMedium_changeEncryption' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'currentPassword', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cipher', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'newPassword', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'newPasswordId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_changeEncryption
'IGuestMouseEvent_getButtons' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMouseEvent_getButtons
'IMachine_getPCIDeviceAssignments' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getPCIDeviceAssignments
'ISystemProperties_getExclusiveHwVirt' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getExclusiveHwVirt
'IGuestSession_close' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_close
'IGuestFileRegisteredEvent_getRegistered' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileRegisteredEvent_getRegistered
'IStorageController_getPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getPortCount
'IDHCPConfig_getForcedOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getForcedOptions
'IRecordingScreenSettings_setFeatures' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'features', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setFeatures
'IUSBDeviceFilter_setRevision' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'revision', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setRevision
'IGuestScreenInfo_getWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getWidth
'IMachineRegisteredEvent_getRegistered' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineRegisteredEvent_getRegistered
'IVirtualBox_setSettingsSecret' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_setSettingsSecret
'IGuestScreenInfo_getGuestMonitorStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getGuestMonitorStatus
'IMedium_checkEncryptionPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_checkEncryptionPassword
'IMachineDebugger_getSingleStep' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getSingleStep
'IBandwidthControl_deleteBandwidthGroup' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthControl_deleteBandwidthGroup
'IGuestSession_getUserHome' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getUserHome
'IMachineDebugger_getLogRelGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogRelGroups
'IMachine_removeSharedFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_removeSharedFolder
'IHostNetworkInterface_getIPAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getIPAddress
'IRecordingScreenSettings_setVideoScalingMethod' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoScalingMethod', type => 'vbox:RecordingVideoScalingMethod', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoScalingMethod
'IMachine_getUSBDeviceFilters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getUSBDeviceFilters
'IDHCPGroupConfig_addCondition' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'inclusive', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:DHCPGroupConditionType', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupConfig_addCondition
'ISnapshot_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getName
'IUnattended_setPostInstallCommand' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'postInstallCommand', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setPostInstallCommand
'INetworkAdapter_getAdapterType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getAdapterType
'INetworkAdapterChangedEvent_getNetworkAdapter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapterChangedEvent_getNetworkAdapter
'IConsole_getPowerButtonHandled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getPowerButtonHandled
'IRecordingScreenSettings_getFilename' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getFilename
'INetworkAdapter_setBridgedInterface' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bridgedInterface', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setBridgedInterface
'IMachineDebugger_setRecompileSupervisor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'recompileSupervisor', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setRecompileSupervisor
'IGuestSession_directoryCreateTemp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'templateName', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'secure', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryCreateTemp
'IFsObjInfo_getFileAttributes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getFileAttributes
'IVRDEServerChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServerChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'INATNetwork_setNeedDhcpServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'needDhcpServer', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setNeedDhcpServer
'IProcess_write' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'handle', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IProcess_write
'IUnattended_setScriptTemplatePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'scriptTemplatePath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setScriptTemplatePath
'IGuest_createSession' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'user', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'domain', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'sessionName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_createSession
'ISystemProperties_getSupportedExportOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedExportOptions
'IUnattended_setUser' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'user', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setUser
'IMouse_getRelativeSupported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouse_getRelativeSupported
'IBIOSSettings_setLogoFadeIn' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logoFadeIn', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setLogoFadeIn
'IMachine_readSavedScreenshotToArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bitmapFormat', type => 'vbox:BitmapFormat', attr => {}),
    ], # end parameters
  }, # end IMachine_readSavedScreenshotToArray
'IMachineDebugger_getVirtualTimeRate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getVirtualTimeRate
'INetworkAdapter_getMACAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getMACAddress
'IConsole_addDiskEncryptionPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'clearOnSuspend', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IConsole_addDiskEncryptionPassword
'ISystemProperties_setLoggingLevel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'loggingLevel', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setLoggingLevel
'ISystemProperties_setDefaultMachineFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultMachineFolder', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setDefaultMachineFolder
'IDHCPServer_getConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'scope', type => 'vbox:DHCPConfigScope', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'slot', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'mayAdd', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getConfig
'INetworkAdapter_getPromiscModePolicy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getPromiscModePolicy
'IBIOSSettings_setIOAPICEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IOAPICEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setIOAPICEnabled
'IFsObjInfo_getUserName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getUserName
'IAppliance_getPasswordIds' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getPasswordIds
'IHostUSBDeviceFilter_getAction' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostUSBDeviceFilter_getAction
'INATNetworkPortForwardEvent_getCreate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getCreate
'IDnDModeChangedEvent_getDndMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDModeChangedEvent_getDndMode
'INetworkAdapter_setProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setProperty
'ISerialPort_getHostMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getHostMode
'IVirtualBox_createUnattendedInstaller' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createUnattendedInstaller
'IMachineDebugger_modifyLogFlags' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settings', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_modifyLogFlags
'IBandwidthGroup_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthGroup_getName
'IUSBDevice_getSerialNumber' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getSerialNumber
'IFsObjInfo_getDeviceNumber' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getDeviceNumber
'ISystemProperties_getDefaultMachineFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultMachineFolder
'IVRDEServer_setAuthTimeout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'authTimeout', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_setAuthTimeout
'IGuestUserStateChangedEvent_getDomain' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestUserStateChangedEvent_getDomain
'IUnattended_setLocale' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'locale', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setLocale
'IMachine_getIOCacheSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getIOCacheSize
'IWebsessionManager_logon' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => 'username', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IWebsessionManager_logon
'ISession_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_setName
'IClipboardModeChangedEvent_getClipboardMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IClipboardModeChangedEvent_getClipboardMode
'IMachine_getSessionPID' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSessionPID
'IMediumRegisteredEvent_getMediumId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumRegisteredEvent_getMediumId
'IGuestSession_copyToGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'sources', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'filters', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_copyToGuest
'IMachine_getBootOrder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'position', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getBootOrder
'IMachine_getParavirtDebug' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getParavirtDebug
'IFsObjInfo_getChangeTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getChangeTime
'IUnattended_getDetectedOSLanguages' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getDetectedOSLanguages
'IProgress_getCancelable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getCancelable
'IMachine_setStorageControllerBootable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bootable', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setStorageControllerBootable
'IFile_setSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IFile_setSize
'IProgress_getOperation' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getOperation
'IGuestSession_fsObjExists' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'followSymlinks', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjExists
'IMachine_exportTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'appliance', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'location', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_exportTo
'IProgress_getCanceled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getCanceled
'IStorageDeviceChangedEvent_getSilent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageDeviceChangedEvent_getSilent
'IConsole_getEmulatedUSB' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getEmulatedUSB
'IMedium_moveTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'location', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_moveTo
'IMachine_lockMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'session', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'lockType', type => 'vbox:LockType', attr => {}),
    ], # end parameters
  }, # end IMachine_lockMachine
'IFramebuffer_getWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getWidth
'IUnattended_getPostInstallCommand' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getPostInstallCommand
'ISharedFolder_getAutoMountPoint' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getAutoMountPoint
'IGuestMonitorInfoChangedEvent_getOutput' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorInfoChangedEvent_getOutput
'IHostNetworkInterface_DHCPRediscover' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_DHCPRediscover
'IUSBController_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBController_getName
'IPerformanceCollector_setupMetrics' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'metricNames', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'objects', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'period', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'count', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IPerformanceCollector_setupMetrics
'ISystemProperties_setFreeDiskSpaceError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'freeDiskSpaceError', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setFreeDiskSpaceError
'IEventSource_eventProcessed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'listener', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'event', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventSource_eventProcessed
'ICloudNetworkGatewayInfo_getMacAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetworkGatewayInfo_getMacAddress
'IUSBDevice_getRevision' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getRevision
'ISystemProperties_getMaxNetworkAdapters' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'chipset', type => 'vbox:ChipsetType', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxNetworkAdapters
'IUSBDevice_getPortPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getPortPath
'IGuestProcessRegisteredEvent_getRegistered' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessRegisteredEvent_getRegistered
'IUSBController_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBController_setName
'ICertificate_getIssuerName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getIssuerName
'ISystemProperties_setWebServiceAuthLibrary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'webServiceAuthLibrary', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setWebServiceAuthLibrary
'ICPUChangedEvent_getAdd' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICPUChangedEvent_getAdd
'INetworkAdapter_setBootPriority' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bootPriority', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setBootPriority
'IMedium_getReadOnly' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getReadOnly
'IStorageController_getBootable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getBootable
'IMouse_putMouseEventAbsolute' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'dz', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'dw', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'buttonState', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IMouse_putMouseEventAbsolute
'IFsObjInfo_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getName
'IMousePointerShapeChangedEvent_getAlpha' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getAlpha
'IMouse_putEventMultiTouch' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'count', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'contacts', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'scanTime', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMouse_putEventMultiTouch
'IStorageController_getMaxPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getMaxPortCount
'ICertificate_queryInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'what', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end ICertificate_queryInfo
'IDataStream_read' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'size', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDataStream_read
'IVFSExplorer_getPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_getPath
'IMedium_getProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getProperties
'IBIOSSettings_setBootMenuMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bootMenuMode', type => 'vbox:BIOSBootMenuMode', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setBootMenuMode
'IPerformanceMetric_getObject' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getObject
'IMachine_getPointingHIDType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getPointingHIDType
'IVirtualBox_getGenericNetworkDrivers' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getGenericNetworkDrivers
'ISystemProperties_getSupportedPortModes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedPortModes
'IBIOSSettings_getACPIEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getACPIEnabled
'IMachine_querySavedScreenshotInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_querySavedScreenshotInfo
'IMousePointerShape_getHotY' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getHotY
'IGuest_getFacilityStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'facility', type => 'vbox:AdditionsFacilityType', attr => {}),
    ], # end parameters
  }, # end IGuest_getFacilityStatus
'IVirtualBox_createDHCPServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createDHCPServer
'IMedium_openForIO' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'writable', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'password', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_openForIO
'ICloudProfile_getProviderId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_getProviderId
'IUnattended_getTimeZone' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getTimeZone
'ISystemProperties_getSupportedUartTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedUartTypes
'IDnDTarget_drop' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'x', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'y', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'defaultAction', type => 'vbox:DnDAction', attr => {}),
      SOAP::Data->new(name => 'allowedActions', type => 'vbox:DnDAction', attr => {}),
      SOAP::Data->new(name => 'formats', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDTarget_drop
'IUSBDeviceStateChangedEvent_getDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceStateChangedEvent_getDevice
'IMachineDebugger_getHWVirtExEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getHWVirtExEnabled
'IRecordingScreenSettings_getAudioBits' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getAudioBits
'IGuestSession_fileCopyFromGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'source', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:FileCopyFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileCopyFromGuest
'ISystemProperties_getFreeDiskSpacePercentError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getFreeDiskSpacePercentError
'INATEngine_getNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getNetwork
'ICloudClient_deleteImage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_deleteImage
'IGuestMouseEvent_getW' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMouseEvent_getW
'INetworkAdapter_setCableConnected' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cableConnected', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setCableConnected
'IUnattended_detectIsoOS' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_detectIsoOS
'IMachineDebugger_dumpGuestStack' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_dumpGuestStack
'IFramebuffer_videoModeSupported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bpp', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_videoModeSupported
'ICertificate_getPublicKeyAlgorithm' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getPublicKeyAlgorithm
'INetworkAdapter_getNATNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getNATNetwork
'ICloudNetworkGatewayInfo_getPublicIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetworkGatewayInfo_getPublicIP
'IHost_getAcceleration3DAvailable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getAcceleration3DAvailable
'IGuestSession_getUser' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getUser
'IVirtualSystemDescription_removeDescriptionByType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:VirtualSystemDescriptionType', attr => {}),
    ], # end parameters
  }, # end IVirtualSystemDescription_removeDescriptionByType
'IDHCPConfig_getDefaultLeaseTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getDefaultLeaseTime
'IVirtualBox_setExtraData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_setExtraData
'IMachine_showConsoleWindow' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_showConsoleWindow
'IMediumFormat_getCapabilities' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumFormat_getCapabilities
'INATNetworkPortForwardEvent_getHostIp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getHostIp
'ICertificate_getRawCertData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getRawCertData
'IGuestMultiTouchEvent_getContactIds' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMultiTouchEvent_getContactIds
'IUnattended_getDetectedOSTypeId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getDetectedOSTypeId
'IDnDTarget_sendData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDTarget_sendData
'ISerialPort_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getEnabled
'ISystemProperties_setDefaultVRDEExtPack' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'defaultVRDEExtPack', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setDefaultVRDEExtPack
'IConsole_getUseHostClipboard' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getUseHostClipboard
'IVirtualBox_findMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'nameOrId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_findMachine
'IMachine_deleteSnapshotAndAllChildren' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_deleteSnapshotAndAllChildren
'IVirtualBoxErrorInfo_getNext' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBoxErrorInfo_getNext
'IVRDEServer_getAllowMultiConnection' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getAllowMultiConnection
'ISession_unlockMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_unlockMachine
'IDHCPGroupCondition_getInclusive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupCondition_getInclusive
'IUSBDeviceFilters_removeDeviceFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'position', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilters_removeDeviceFilter
'INATEngine_setDNSProxy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'DNSProxy', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATEngine_setDNSProxy
'IMachine_getAutostopType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAutostopType
'IConsole_clearAllDiskEncryptionPasswords' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_clearAllDiskEncryptionPasswords
'IMachineDebugger_setLogEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'logEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_setLogEnabled
'IMachine_getRTCUseUTC' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getRTCUseUTC
'IFsObjInfo_getGenerationId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getGenerationId
'IDHCPConfig_removeAllOptions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_removeAllOptions
'IProcess_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProcess_getName
'IMediumIO_initializePartitionTable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'vbox:PartitionTableType', attr => {}),
      SOAP::Data->new(name => 'wholeDiskInOneEntry', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMediumIO_initializePartitionTable
'IMedium_refreshState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_refreshState
'IUSBDeviceFilter_setActive' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'active', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setActive
'IGuestAdditionsStatusChangedEvent_getTimestamp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestAdditionsStatusChangedEvent_getTimestamp
'IFramebuffer_notifyChange' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'xOrigin', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'yOrigin', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_notifyChange
'IMachine_readSavedThumbnailToArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bitmapFormat', type => 'vbox:BitmapFormat', attr => {}),
    ], # end parameters
  }, # end IMachine_readSavedThumbnailToArray
'IHostNetworkInterface_getWireless' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getWireless
'ICloudClient_startCloudNetworkGateway' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'network', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'sshPublicKey', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_startCloudNetworkGateway
'IMachine_setTracingConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'tracingConfig', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setTracingConfig
'IMachine_getCPUProfile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUProfile
'IHost_getProcessorCPUIDLeaf' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'leaf', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'subLeaf', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorCPUIDLeaf
'IMachine_setClipboardFileTransfersEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'clipboardFileTransfersEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setClipboardFileTransfersEnabled
'IParallelPort_setEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IParallelPort_setEnabled
'IDHCPGroupConfig_setName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupConfig_setName
'IMachineDebugger_getStats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'pattern', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'withDescriptions', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getStats
'IMachine_getBandwidthControl' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getBandwidthControl
'IDisplay_getGuestScreenLayout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDisplay_getGuestScreenLayout
'IMachine_getEffectiveParavirtProvider' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getEffectiveParavirtProvider
'IMedium_getChildren' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getChildren
'IMousePointerShapeChangedEvent_getYhot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getYhot
'IMachine_setGuestProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setGuestProperty
'IMedium_close' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_close
'ICloudNetwork_getNetworkId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_getNetworkId
'IUSBDevice_getRemote' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getRemote
'IDHCPServer_getLowerIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getLowerIP
'IBandwidthGroup_setMaxBytesPerSec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'maxBytesPerSec', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IBandwidthGroup_setMaxBytesPerSec
'IVirtualBox_checkFirmwarePresent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'firmwareType', type => 'vbox:FirmwareType', attr => {}),
      SOAP::Data->new(name => 'version', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_checkFirmwarePresent
'IGuestSession_fileExists' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'followSymlinks', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileExists
'IUSBDeviceFilter_getProduct' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getProduct
'IPerformanceMetric_getPeriod' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getPeriod
'IMachine_getCPUStatus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpu', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUStatus
'IUnattended_getPassword' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getPassword
'IFramebuffer_getHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFramebuffer_getHeight
'IMachine_deleteGuestProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_deleteGuestProperty
'IFile_writeAt' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'data', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFile_writeAt
'IVirtualBoxErrorInfo_getComponent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBoxErrorInfo_getComponent
'IGuestFileWriteEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileWriteEvent_getMidlDoesNotLikeEmptyInterfaces
'IParallelPort_setIRQ' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IRQ', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IParallelPort_setIRQ
'IMedium_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getName
'IPerformanceMetric_getMetricName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getMetricName
'IDHCPServer_getNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getNetworkName
'INetworkAdapter_getCableConnected' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getCableConnected
'IGuestMonitorChangedEvent_getWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorChangedEvent_getWidth
'IEmulatedUSB_webcamDetach' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEmulatedUSB_webcamDetach
'IVirtualBox_openMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'settingsFile', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_openMachine
'IMousePointerShape_getShape' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getShape
'IMachine_getAccessError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getAccessError
'IHost_getProcessorSpeed' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorSpeed
'ISystemProperties_getMaxGuestRAM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxGuestRAM
'IMachineStateChangedEvent_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineStateChangedEvent_getState
'IVirtualBox_getHomeFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getHomeFolder
'ISnapshot_getTimeStamp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getTimeStamp
'ICloudNetwork_getNetworkName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_getNetworkName
'IAudioAdapter_setEnabledOut' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabledOut', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setEnabledOut
'IFormValue_getVisible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getVisible
'IMachine_setTeleporterAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'teleporterAddress', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setTeleporterAddress
'IMediumIO_getMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumIO_getMedium
'IBandwidthControl_getBandwidthGroup' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthControl_getBandwidthGroup
'ISystemProperties_getMinPortCountForStorageBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bus', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMinPortCountForStorageBus
'INATEngine_getTFTPBootFile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getTFTPBootFile
'IFormValue_getLabel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getLabel
'IHostPCIDevicePlugEvent_getAttachment' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostPCIDevicePlugEvent_getAttachment
'IVirtualBox_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getEventSource
'ISerialPort_getSlot' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getSlot
'IConsole_getMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getMachine
'IRangedIntegerFormValue_getInteger' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRangedIntegerFormValue_getInteger
'IBIOSSettings_setPXEDebugEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'PXEDebugEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setPXEDebugEnabled
'ISession_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISession_getType
'IMachine_setHPETEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'HPETEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setHPETEnabled
'IRecordingScreenSettings_getVideoCodec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoCodec
'IGuestMouseEvent_getZ' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMouseEvent_getZ
'ICertificate_getVersionNumber' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getVersionNumber
'ICloudProfile_getProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_getProperty
'ICertificate_getKeyUsage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getKeyUsage
'IStorageDeviceChangedEvent_getStorageDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageDeviceChangedEvent_getStorageDevice
'IHost_removeHostOnlyNetworkInterface' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_removeHostOnlyNetworkInterface
'IHostNetworkInterface_enableDynamicIPConfig' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_enableDynamicIPConfig
'IGuestMonitorChangedEvent_getOriginY' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMonitorChangedEvent_getOriginY
'IProgress_getOperationDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getOperationDescription
'IConsole_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getEventSource
'IAudioAdapter_getProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getProperty
'ISystemProperties_getSupportedNetworkAttachmentTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedNetworkAttachmentTypes
'IMachine_unregister' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cleanupMode', type => 'vbox:CleanupMode', attr => {}),
    ], # end parameters
  }, # end IMachine_unregister
'IFile_readAt' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'offset', type => 'xsd:long', attr => {}),
      SOAP::Data->new(name => 'toRead', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IFile_readAt
'IVirtualBox_getDHCPServers' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getDHCPServers
'IKeyboardLedsChangedEvent_getScrollLock' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IKeyboardLedsChangedEvent_getScrollLock
'ISharedFolder_getHostPath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getHostPath
'ISystemProperties_getRawModeSupported' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getRawModeSupported
'IRecordingScreenSettings_getMaxFileSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getMaxFileSize
'IUSBDeviceStateChangedEvent_getError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceStateChangedEvent_getError
'IMachine_setExtraData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setExtraData
'IGuestScreenInfo_getBitsPerPixel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getBitsPerPixel
'IParallelPortChangedEvent_getParallelPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPortChangedEvent_getParallelPort
'IMachineDebugger_getCSAMEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getCSAMEnabled
'IConsole_resume' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_resume
'ISystemProperties_getSupportedVMProcPriorities' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedVMProcPriorities
'IMediumFormat_describeFileExtensions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumFormat_describeFileExtensions
'IMachine_setIcon' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'icon', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setIcon
'IMachine_getMediumAttachmentsOfController' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getMediumAttachmentsOfController
'IVirtualBox_createCloudNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createCloudNetwork
'IParallelPort_getIOBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IParallelPort_getIOBase
'IMediumIO_getWritable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumIO_getWritable
'IExtraDataCanChangeEvent_getValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IExtraDataCanChangeEvent_getValue
'ICertificate_getIssuerUniqueIdentifier' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getIssuerUniqueIdentifier
'ICloudClient_terminateInstance' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uid', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudClient_terminateInstance
'IMachine_setCPUIDPortabilityLevel' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'CPUIDPortabilityLevel', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUIDPortabilityLevel
'ICloudProvider_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getName
'IFsObjInfo_getBirthTime' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFsObjInfo_getBirthTime
'IGuestScreenInfo_getScreenId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getScreenId
'IFormValue_getGeneration' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFormValue_getGeneration
'IFile_getFilename' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getFilename
'INetworkAdapter_setNATNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'NATNetwork', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setNATNetwork
'IProgress_getInitiator' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getInitiator
'IMachine_getHWVirtExProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'vbox:HWVirtExPropertyType', attr => {}),
    ], # end parameters
  }, # end IMachine_getHWVirtExProperty
'IStorageController_setPortCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'portCount', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IStorageController_setPortCount
'IGuestSession_directoryRemove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryRemove
'IGuestSession_setTimeout' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'timeout', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_setTimeout
'INATEngine_setDNSPassDomain' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'DNSPassDomain', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end INATEngine_setDNSPassDomain
'IGuestSession_getCurrentDirectory' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getCurrentDirectory
'IProcess_read' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'handle', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'toRead', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IProcess_read
'IGuestProcess_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcess_getMidlDoesNotLikeEmptyInterfaces
'ISharedFolder_getAutoMount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_getAutoMount
'IProgress_getOperationWeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getOperationWeight
'IBIOSSettings_getNonVolatileStorageFile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getNonVolatileStorageFile
'IMachine_createSharedFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'hostPath', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'writable', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'automount', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'autoMountPoint', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_createSharedFolder
'IBandwidthControl_getAllBandwidthGroups' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBandwidthControl_getAllBandwidthGroups
'INetworkAdapter_getProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getProperty
'IMachine_setParavirtDebug' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'paravirtDebug', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setParavirtDebug
'IDHCPGroupConfig_removeAllConditions' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPGroupConfig_removeAllConditions
'IEventSource_createListener' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEventSource_createListener
'IMachine_getPageFusionEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getPageFusionEnabled
'IGuestKeyboardEvent_getScancodes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestKeyboardEvent_getScancodes
'IUnattended_getMachine' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getMachine
'IVirtualBox_findNATNetworkByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'networkName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_findNATNetworkByName
'INATEngine_getDNSProxy' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getDNSProxy
'IShowWindowEvent_setWinId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'winId', type => 'xsd:long', attr => {}),
    ], # end parameters
  }, # end IShowWindowEvent_setWinId
'INATNetwork_getIPv6Enabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getIPv6Enabled
'IUSBDevice_getProductId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getProductId
'ISystemProperties_getWebServiceAuthLibrary' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getWebServiceAuthLibrary
'IVFSExplorer_remove' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_remove
'ICloudProfile_setProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProfile_setProperty
'IPCIAddress_getDevFunction' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_getDevFunction
'IMouseCapabilityChangedEvent_getSupportsMultiTouch' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouseCapabilityChangedEvent_getSupportsMultiTouch
'IAudioAdapter_setEnabledIn' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'enabledIn', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setEnabledIn
'IVirtualBox_removeDHCPServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'server', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_removeDHCPServer
'IRecordingScreenSettings_setAudioBits' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioBits', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setAudioBits
'IWebsessionManager_getSessionObject' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => 'refIVirtualBox', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IWebsessionManager_getSessionObject
'ICloudNetwork_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudNetwork_getEnabled
'IMachine_attachDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:DeviceType', attr => {}),
      SOAP::Data->new(name => 'medium', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_attachDevice
'IVirtualBox_getMachines' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getMachines
'IPerformanceCollector_disableMetrics' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'metricNames', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'objects', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceCollector_disableMetrics
'IMachine_getStateFilePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getStateFilePath
'IAudioAdapter_setAudioDriver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'audioDriver', type => 'vbox:AudioDriverType', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_setAudioDriver
'IAppliance_createVFSExplorer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'URI', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_createVFSExplorer
'IHost_findHostNetworkInterfacesOfType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:HostNetworkInterfaceType', attr => {}),
    ], # end parameters
  }, # end IHost_findHostNetworkInterfacesOfType
'IDisplay_detachFramebuffer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDisplay_detachFramebuffer
'IGuest_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getEventSource
'IHostNetworkInterface_getDHCPEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getDHCPEnabled
'IBIOSSettings_getLogoFadeOut' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getLogoFadeOut
'ISystemProperties_getSupportedClipboardModes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedClipboardModes
'IConsole_getRemoteUSBDevices' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getRemoteUSBDevices
'IVFSExplorer_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_getType
'IGuestSession_copyFromGuest' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'sources', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'filters', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'destination', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_copyFromGuest
'IGuestSession_getDomain' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getDomain
'IMachine_attachDeviceWithoutMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:DeviceType', attr => {}),
    ], # end parameters
  }, # end IMachine_attachDeviceWithoutMedium
'INATEngine_setNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'network', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_setNetwork
'INATRedirectEvent_getGuestIP' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATRedirectEvent_getGuestIP
'IDnDBase_removeFormats' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'formats', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDnDBase_removeFormats
'IMouse_getNeedsHostCursor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouse_getNeedsHostCursor
'IMachine_temporaryEjectDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'controllerPort', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'device', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'temporaryEject', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_temporaryEjectDevice
'IRecordingScreenSettings_setFilename' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'filename', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setFilename
'IUSBDeviceFilter_getName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_getName
'INATNetwork_setLoopbackIp6' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'loopbackIp6', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end INATNetwork_setLoopbackIp6
'IMachineDebugger_getLogRelFlags' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogRelFlags
'IBIOSSettings_setACPIEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'ACPIEnabled', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_setACPIEnabled
'IVRDEServer_getVRDEProperty' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getVRDEProperty
'ISerialPort_setUartType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'uartType', type => 'vbox:UartType', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setUartType
'IMediumConfigChangedEvent_getMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumConfigChangedEvent_getMedium
'IMachine_setCPUProfile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'CPUProfile', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setCPUProfile
'IUnattended_setProductKey' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'productKey', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setProductKey
'IVFSExplorer_update' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVFSExplorer_update
'IGuestSession_environmentGetBaseVariable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_environmentGetBaseVariable
'IUSBController_getType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBController_getType
'IFile_getOffset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IFile_getOffset
'IMachineDebugger_getLogDbgDestinations' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getLogDbgDestinations
'IMachine_setRTCUseUTC' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'RTCUseUTC', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IMachine_setRTCUseUTC
'ISnapshotChangedEvent_getMidlDoesNotLikeEmptyInterfaces' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshotChangedEvent_getMidlDoesNotLikeEmptyInterfaces
'ISystemProperties_getSupportedAudioControllerTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedAudioControllerTypes
'ISharedFolder_setWritable' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'writable', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_setWritable
'ICertificate_getFriendlyName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getFriendlyName
'IBIOSSettings_getTimeOffset' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IBIOSSettings_getTimeOffset
'IMachine_moveTo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'folder', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_moveTo
'ISharedFolder_setAutoMountPoint' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autoMountPoint', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISharedFolder_setAutoMountPoint
'IMachine_setAutostopType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autostopType', type => 'vbox:AutostopType', attr => {}),
    ], # end parameters
  }, # end IMachine_setAutostopType
'IGuestSession_directoryCreate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'mode', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:DirectoryCreateFlag', attr => {}),
    ], # end parameters
  }, # end IGuestSession_directoryCreate
'IHost_insertUSBDeviceFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'position', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'filter', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_insertUSBDeviceFilter
'IUSBController_setType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'type', type => 'vbox:USBControllerType', attr => {}),
    ], # end parameters
  }, # end IUSBController_setType
'IMachine_getCurrentStateModified' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCurrentStateModified
'ISerialPort_setIOBase' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'IOBase', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end ISerialPort_setIOBase
'INATEngine_getTFTPNextServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATEngine_getTFTPNextServer
'IMousePointerShape_getHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getHeight
'IStorageControllerChangedEvent_getMachinId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageControllerChangedEvent_getMachinId
'IMachineDebugger_getRecompileSupervisor' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getRecompileSupervisor
'ISystemProperties_setAutostartDatabasePath' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autostartDatabasePath', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setAutostartDatabasePath
'INetworkAdapter_getAttachmentType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getAttachmentType
'IGuestSession_waitForArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'waitFor', type => 'vbox:GuestSessionWaitForFlag', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IGuestSession_waitForArray
'IMachineDebugger_getCPULoad' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'cpuId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_getCPULoad
'IAppliance_getCertificate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_getCertificate
'IGuest_getAdditionsRevision' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuest_getAdditionsRevision
'IConsole_getDisplay' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getDisplay
'IGuestSession_fileQuerySize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'followSymlinks', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fileQuerySize
'IVirtualBox_getPackageType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getPackageType
'IHostNetworkInterface_getHardwareAddress' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHostNetworkInterface_getHardwareAddress
'IGuestSession_getEventSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_getEventSource
'IExtraDataCanChangeEvent_getKey' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IExtraDataCanChangeEvent_getKey
'IMedium_getParent' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getParent
'IMousePointerShape_getWidth' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getWidth
'ISnapshot_getOnline' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISnapshot_getOnline
'IPerformanceCollector_getMetricNames' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceCollector_getMetricNames
'IMachine_getHPETEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getHPETEnabled
'IPCIAddress_setDevFunction' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'devFunction', type => 'xsd:short', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_setDevFunction
'ICertificate_getSelfSigned' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSelfSigned
'IHost_addUSBDeviceSource' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'backend', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'address', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'propertyNames', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'propertyValues', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_addUSBDeviceSource
'IUSBDeviceFilter_setProductId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'productId', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setProductId
'IEmulatedUSB_getWebcams' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IEmulatedUSB_getWebcams
'ICloudProvider_getProfileByName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'profileName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getProfileByName
'INATNetwork_getNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getNetwork
'IVirtualBox_getInternalNetworks' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getInternalNetworks
'IAppliance_write' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'options', type => 'vbox:ExportOptions', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAppliance_write
'ICloudProvider_getId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getId
'IMachine_getSnapshotFolder' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getSnapshotFolder
'IMachine_setAutostartDelay' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'autostartDelay', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_setAutostartDelay
'IStorageController_getControllerType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStorageController_getControllerType
'IUnattended_getDetectedOSVersion' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getDetectedOSVersion
'IGuestFileSizeChangedEvent_getNewSize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestFileSizeChangedEvent_getNewSize
'IUnattended_getDetectedOSHints' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_getDetectedOSHints
'IGuestUserStateChangedEvent_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestUserStateChangedEvent_getState
'ISystemProperties_getMaxInstancesOfStorageBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'chipset', type => 'vbox:ChipsetType', attr => {}),
      SOAP::Data->new(name => 'bus', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxInstancesOfStorageBus
'IStateChangedEvent_getState' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IStateChangedEvent_getState
'ISystemProperties_getSupportedUSBControllerTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedUSBControllerTypes
'IGuestSession_symlinkExists' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'symlink', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_symlinkExists
'IMediumRegisteredEvent_getMediumType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMediumRegisteredEvent_getMediumType
'ISystemProperties_getMinGuestRAM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMinGuestRAM
'IGuestScreenInfo_getOriginY' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestScreenInfo_getOriginY
'INATNetwork_getNeedDhcpServer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getNeedDhcpServer
'IGuestSession_fsObjQueryInfo' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'followSymlinks', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjQueryInfo
'IGuestMouseEvent_getY' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestMouseEvent_getY
'IRecordingScreenSettings_getVideoRate' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_getVideoRate
'IGuestProcessStateChangedEvent_getError' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestProcessStateChangedEvent_getError
'ISystemProperties_getDefaultVRDEExtPack' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultVRDEExtPack
'IUSBDevice_getPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getPort
'IAudioAdapter_getAudioDriver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IAudioAdapter_getAudioDriver
'IDHCPConfig_getOption' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'option', type => 'vbox:DHCPOption', attr => {}),
    ], # end parameters
  }, # end IDHCPConfig_getOption
'IHost_getMemorySize' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getMemorySize
'INATNetwork_getEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getEnabled
'IUnattended_setCountry' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'country', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setCountry
'IUSBDeviceFilter_setPort' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'port', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilter_setPort
'IDnDSource_dragIsPending' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'screenId', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IDnDSource_dragIsPending
'IPerformanceMetric_getMinimumValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPerformanceMetric_getMinimumValue
'INetworkAdapter_setInternalNetwork' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'internalNetwork', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setInternalNetwork
'IConsole_detachUSBDevice' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'id', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_detachUSBDevice
'ISystemProperties_setFreeDiskSpacePercentWarning' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'freeDiskSpacePercentWarning', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_setFreeDiskSpacePercentWarning
'ISystemProperties_getSupportedFirmwareTypes' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getSupportedFirmwareTypes
'IUSBDeviceFilters_createDeviceFilter' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'name', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDeviceFilters_createDeviceFilter
'IUnattended_reconfigureVM' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_reconfigureVM
'INATNetwork_getPortForwardRules6' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetwork_getPortForwardRules6
'IMachine_removeCPUIDLeaf' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'idx', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'idxSub', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IMachine_removeCPUIDLeaf
'IMedium_compact' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_compact
'IGuestSession_processCreateEx' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'executable', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'arguments', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'environmentChanges', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'vbox:ProcessCreateFlag', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'priority', type => 'vbox:ProcessPriority', attr => {}),
      SOAP::Data->new(name => 'affinity', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IGuestSession_processCreateEx
'IDHCPIndividualConfig_getMachineId' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPIndividualConfig_getMachineId
'ISerialPort_getIRQ' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISerialPort_getIRQ
'IMedium_getEncryptionSettings' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_getEncryptionSettings
'IMouse_getPointerShape' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMouse_getPointerShape
'INetworkAdapter_setTraceFile' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'traceFile', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_setTraceFile
'IMousePointerShapeChangedEvent_getHeight' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getHeight
'IMousePointerShape_getHotX' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShape_getHotX
'ISystemProperties_getFreeDiskSpaceWarning' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getFreeDiskSpaceWarning
'ISystemProperties_getMaxDevicesPerPortForStorageBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'bus', type => 'vbox:StorageBus', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getMaxDevicesPerPortForStorageBus
'INATNetworkSettingEvent_getAdvertiseDefaultIPv6RouteEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkSettingEvent_getAdvertiseDefaultIPv6RouteEnabled
'ICertificate_getSubjectUniqueIdentifier' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICertificate_getSubjectUniqueIdentifier
'IKeyboard_putScancode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'scancode', type => 'xsd:int', attr => {}),
    ], # end parameters
  }, # end IKeyboard_putScancode
'IHost_getProcessorOnlineCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_getProcessorOnlineCount
'INATNetworkPortForwardEvent_getGuestIp' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INATNetworkPortForwardEvent_getGuestIp
'IMachine_getCPUHotPlugEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getCPUHotPlugEnabled
'IGuestAdditionsStatusChangedEvent_getFacility' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestAdditionsStatusChangedEvent_getFacility
'IMachine_getIcon' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getIcon
'ICloudProvider_getSupportedPropertyNames' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProvider_getSupportedPropertyNames
'INetworkAdapter_getBandwidthGroup' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getBandwidthGroup
'IUnattended_setLanguage' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'language', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUnattended_setLanguage
'IVirtualBox_getAPIRevision' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_getAPIRevision
'IProcess_waitForArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'waitFor', type => 'vbox:ProcessWaitForFlag', attr => {}),
      SOAP::Data->new(name => 'timeoutMS', type => 'xsd:unsignedInt', attr => {}),
    ], # end parameters
  }, # end IProcess_waitForArray
'IMachine_getExtraData' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'key', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getExtraData
'IMachine_getTracingEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getTracingEnabled
'IDisplay_setVideoModeHint' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'display', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'enabled', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'changeOrigin', type => 'xsd:boolean', attr => {}),
      SOAP::Data->new(name => 'originX', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'originY', type => 'xsd:int', attr => {}),
      SOAP::Data->new(name => 'width', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'height', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'bitsPerPixel', type => 'xsd:unsignedInt', attr => {}),
      SOAP::Data->new(name => 'notify', type => 'xsd:boolean', attr => {}),
    ], # end parameters
  }, # end IDisplay_setVideoModeHint
'IGuestSession_fsObjRemoveArray' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'path', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IGuestSession_fsObjRemoveArray
'IMachine_getClipboardMode' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getClipboardMode
'IConsole_getUSBDevices' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IConsole_getUSBDevices
'ISystemProperties_getDefaultAudioDriver' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ISystemProperties_getDefaultAudioDriver
'IMachineDebugger_dumpGuestCore' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'filename', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'compression', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachineDebugger_dumpGuestCore
'INetworkAdapter_getTraceEnabled' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getTraceEnabled
'IMachine_setGuestPropertyValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'value', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_setGuestPropertyValue
'IRecordingScreenSettings_setVideoCodec' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'videoCodec', type => 'vbox:RecordingVideoCodec', attr => {}),
    ], # end parameters
  }, # end IRecordingScreenSettings_setVideoCodec
'IPCIAddress_getBus' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IPCIAddress_getBus
'IDHCPServer_getGroupConfigs' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IDHCPServer_getGroupConfigs
'IProgress_getDescription' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getDescription
'IHost_createHostOnlyNetworkInterface' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IHost_createHostOnlyNetworkInterface
'IVRDEServer_getAuthType' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IVRDEServer_getAuthType
'IMachine_getGuestPropertyValue' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'property', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_getGuestPropertyValue
'IMedium_setProperties' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'names', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'values', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMedium_setProperties
'IUSBDevice_getManufacturer' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IUSBDevice_getManufacturer
'IMachine_applyDefaults' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'flags', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMachine_applyDefaults
'IVirtualBox_createMedium' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'format', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'location', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'accessMode', type => 'vbox:AccessMode', attr => {}),
      SOAP::Data->new(name => 'aDeviceTypeType', type => 'vbox:DeviceType', attr => {}),
    ], # end parameters
  }, # end IVirtualBox_createMedium
'ICloudProviderManager_getProviderByShortName' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
      SOAP::Data->new(name => 'providerName', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end ICloudProviderManager_getProviderByShortName
'INetworkAdapter_getHostOnlyInterface' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end INetworkAdapter_getHostOnlyInterface
'IMousePointerShapeChangedEvent_getVisible' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IMousePointerShapeChangedEvent_getVisible
'IProgress_getOperationCount' => {
    endpoint => 'http://localhost:18083/',
    soapaction => '',
    namespace => 'http://www.virtualbox.org/',
    parameters => [
      SOAP::Data->new(name => '_this', type => 'xsd:string', attr => {}),
    ], # end parameters
  }, # end IProgress_getOperationCount
); # end my %methods

use SOAP::Lite;
use Exporter;
use Carp ();
%methods = (%methods, %custom_methods);
use vars qw(@ISA $AUTOLOAD @EXPORT_OK %EXPORT_TAGS);
@ISA = qw(Exporter SOAP::Lite);
@EXPORT_OK = (keys %methods);
push @EXPORT_OK, qw($endpoint $fault);
%EXPORT_TAGS = ('all' => [@EXPORT_OK]);


sub _call {
    my ($self, $method) = (shift, shift);
    my $name = UNIVERSAL::isa($method => 'SOAP::Data') ? $method->name : $method;
    my %method = %{$methods{$name}};
    $self->soapversion('1.1'); # Lets insist on SOAP 1.1
    $self->proxy($method{endpoint} || Carp::croak "No server address (proxy) specified")
        unless $self->proxy;
    my @templates = @{$method{parameters}};
    my @parameters = ();
    foreach my $param (@_) {
        if (@templates) {
            my $template = shift @templates;
            my ($prefix,$typename) = SOAP::Utils::splitqname($template->type);
            my $method = 'as_'.$typename;
            # TODO - if can('as_'.$typename) {...}
            my $result = $self->serializer->$method($param, $template->name, $template->type, $template->attr);
            push(@parameters, $template->value($result->[2]));
        }
        else {
            push(@parameters, $param);
        }
    }
    $self->on_fault($fault);
    $self->endpoint($endpoint)
       ->ns($method{namespace})
       ->on_action(sub{qq!"$method{soapaction}"!});
  $self->serializer->register_ns("http://www.virtualbox.org/","vbox");
  $self->serializer->register_ns("urn:vbox","interface");
    my $som = $self->SUPER::call($method => @parameters);
    if ($self->want_som) {
        return $som;
    }
    UNIVERSAL::isa($som => 'SOAP::SOM') ? wantarray ? $som->paramsall : $som->result : $som;
}

sub BEGIN {
    no strict 'refs';
    for my $method (qw(want_som)) {
        my $field = '_' . $method;
        *$method = sub {
            my $self = shift->new;
            @_ ? ($self->{$field} = shift, return $self) : return $self->{$field};
        }
    }
}
no strict 'refs';
for my $method (@EXPORT_OK) {
    my %method = %{$methods{$method}};
    *$method = sub {
        my $self = UNIVERSAL::isa($_[0] => __PACKAGE__)
            ? ref $_[0]
                ? shift # OBJECT
                # CLASS, either get self or create new and assign to self
                : (shift->self || __PACKAGE__->self(__PACKAGE__->new))
            # function call, either get self or create new and assign to self
            : (__PACKAGE__->self || __PACKAGE__->self(__PACKAGE__->new));
        $self->_call($method, @_);
    }
}

sub AUTOLOAD {
    my $method = substr($AUTOLOAD, rindex($AUTOLOAD, '::') + 2);
    return if $method eq 'DESTROY' || $method eq 'want_som';
    die "Unrecognized method '$method'. List of available method(s): @EXPORT_OK\n";
}

1;
